import matplotlib.pyplot as plt

__all__ = ['seeing_plot', 'plot_predict_ingress_egress']


def seeing_plot(raw_radius, raw_counts, binned_radius, binned_counts, HWHM,
                plot_title='', file_name='', gap=6, annulus_width=13, radius=None,
                figsize=(20, 10)):
    """
    Show a seeing plot for data from an image with radius on the x axis and counts (ADU) on the y axis.

    Parameters
    ----------
    raw_radius : array
        the distance of each pixel from the object of interest

    raw_counts : array
        the counts of each pixel

    binned_radius : array
        pixels grouped by distance from object of interest

    binned_counts : array
        average counts of each group of pixels

    HWHM : number
        half width half max, 1/2 * FWHM

    plot_title : optional, string
        title of plot

    file_name : optional, string
        if entered, file will save as png with this name

    gap : number
        the distance between the aperture and the inner annulus

    annulus_width : number
        the distance between the inner and outer annulus

    figsize : tuple of int, optional
        Size of figure.

    Returns
    -------

    `matplotlib.pyplot.figure`
        The figure object containing the seeing plot.
    """
    if radius is None:
        radius = HWHM * 4

    fig = plt.figure(figsize=figsize)
    plt.grid(True)
    inner_annulus = radius + gap
    outer_annulus = inner_annulus + annulus_width

    # plot the raw radius and raw counts
    plt.plot(raw_radius, raw_counts, linestyle='none',
             marker="s", markerfacecolor='none', color='blue')

    # plot the binned radius and binned counts
    plt.plot(binned_radius, binned_counts, color='magenta', linewidth='1.0')

    # draw vertical line at HWHM and label it
    plt.vlines(HWHM, -0.2, 1.2, linestyle=(0, (5, 10)), color='#00cc00')
    plt.annotate(f"HWHM {HWHM:2.1f}", (HWHM, -0.25),
                 color='#00cc00', horizontalalignment='center')

    # label axis
    plt.xlabel('Radius (pixels)')
    plt.ylabel('ADU')

    # draw vertical line at the radius and label it
    plt.vlines(radius, -0.2, binned_counts[0], color='red')
    plt.annotate(f"Radius {radius:2.1f}", (radius, -0.25),
                 color='red', horizontalalignment='center')
    plt.hlines(binned_counts[0], binned_counts[0], radius, color='red')

    # label the source
    plt.annotate(
        'SOURCE', (radius, binned_counts[0] + 0.02),
        color='red', horizontalalignment='center')

    # draw vertical lines at the background and label it
    plt.vlines(inner_annulus, -0.2, binned_counts[0], color='red')
    plt.vlines(outer_annulus, -0.2, binned_counts[0], color='red')
    plt.hlines(binned_counts[0], inner_annulus, outer_annulus, color='red')
    plt.annotate('BACKGROUND', (inner_annulus,
                                binned_counts[0] + 0.02), color='red')
    plt.annotate(f"Back> {inner_annulus:2.1f}",
                 (inner_annulus, -0.25), color='red', horizontalalignment='center')
    plt.annotate(f"<Back {outer_annulus:2.1f}",
                 (outer_annulus, -0.25), color='red', horizontalalignment='center')

    # title the plot
    title_string = [f"{plot_title}", f"FWHM:{HWHM*2:.1f} pixels"]
    plt.title('\n'.join(title_string))

    # save plot as png
    if file_name:
        safe_name = file_name.replace(" ", "-")
        plt.savefig(f"{safe_name + '-seeing-profile'}.png")
    return fig


def plot_predict_ingress_egress(ingress_time, egress_time, end_line=1,
                                ingress_x_pos=1, egress_x_pos=1, labels_y_pos=1):
    """
    Parameters
    ----------
    ingress_time : float
        the beginning of an exoplanet transit

    egress_time : float
        the end of an exoplanet transit

    end_line : float
        offset to move the vertical lines

    ingress_x_pos : float
        offset to center ingress label

    egress_x_pos : float
        offset to center egress label

    labels_y_pos : float
        offset to move ingress and egress labels

    Returns
    -------

    None
        Directly adds lines and labels to the current plot.
    """
    ymin, ymax = plt.ylim()

    # create a vertical line at the ingress time and label it
    plt.vlines(ingress_time, ymin - end_line, ymax,
               linestyle=(0, (5, 10)), color='red')
    plt.annotate("Predicted Ingress", (ingress_time - ingress_x_pos,
                                       ymin - labels_y_pos), color='red')

    # create a vertical line at the egress time and label it
    plt.vlines(egress_time, ymin - end_line, ymax,
               linestyle=(0, (5, 10)), color='red')
    plt.annotate("Predicted Egress", (egress_time - egress_x_pos,
                                      ymin - labels_y_pos), fontsize=10, color='red')



import numpy as np

from astropy.stats import sigma_clipped_stats
from astropy.modeling.models import Const2D, Gaussian2D
from astropy.modeling.fitting import LevMarLSQFitter

from photutils import DAOStarFinder
from photutils.utils._moments import _moments_central, _moments
from photutils.morphology import data_properties
from astropy.nddata.utils import Cutout2D
from astropy.stats import gaussian_sigma_to_fwhm
from astropy import units as u

__all__ = ['source_detection', 'compute_fwhm']


def _fit_2dgaussian(data):
    """
    Fit a 2D Gaussian to data.

    Written as a replacement for functionality that was removed from
    photutils.

    This function will be kept private so we don't have to support it.

    Copy/pasted from
    https://github.com/astropy/photutils/pull/1064/files#diff-9e64908ff7ac552845b4831870a749f397d73c681d218267bd9087c7757e6dd4R285

    Parameters
    ----------
    data : array-like
        The 2D array of data to fit.

    Returns
    -------
    gfit : `astropy.modeling.Model`
        The best-fit 2D Gaussian model.
    """
    props = data_properties(data - np.min(data))

    init_const = 0.  # subtracted data minimum above
    init_amplitude = np.ptp(data)

    g_init = (Const2D(init_const)
                  + Gaussian2D(amplitude=init_amplitude,
                               x_mean=props.xcentroid,
                               y_mean=props.ycentroid,
                               x_stddev=props.semimajor_sigma.value,
                               y_stddev=props.semiminor_sigma.value,
                               theta=props.orientation.value))

    fitter = LevMarLSQFitter()
    y, x = np.indices(data.shape)
    gfit = fitter(g_init, x, y, data)

    return gfit


def compute_fwhm(ccd, sources, fwhm_estimate=5,
                 x_column='xcenter', y_column='ycenter',
                 fit=True,
                 sky_per_pix_avg=0):
    """
    Computes the FWHM in both x and y directions of sources in an image.

    Parameters
    ----------

    ccd : `astropy.nddata.CCDData`
        The CCD Image array.

    sources : `astropy.table.Table`
        An astropy table of the positions of sources in the image.

    fwhm_estimate : float, optional
        The initial guess for the FWHM of the sources in the image.

    x_column : str, optional
        The name of the column in `sources` that contains the x positions
        of the sources.

    y_column : str, optional
        The name of the column in `sources` that contains the y positions
        of the sources.

    fit : bool, optional
        If ``True``, fit a 2D Gaussian to each source to estimate its FWHM. If
        ``False``, estimate the FWHM of each source by computing the second
        moments of its light distribution using photutils.

    sky_per_pix_avg : float or array-like of float, optional
        Sky background to subtract before centroiding.

    Returns
    -------

    fwhm_x, fwhm_y : array-like of float
        The FWHM of each source in the x and y directions.

    """

    fwhm_x = []
    fwhm_y = []
    for source in sources:
        x = source[x_column]
        y = source[y_column]
        sky = sky_per_pix_avg
        # Cutout2D needs no units on the center position, so remove unit
        # if it is present.
        try:
            x = x.value
            y = y.value
            sky = sky.value
        except AttributeError:
            pass

        cutout = Cutout2D(ccd.data, (x, y), 5 * fwhm_estimate)
        if fit:
            fit = _fit_2dgaussian(cutout.data)
            fwhm_x.append(gaussian_sigma_to_fwhm * fit.x_stddev_1)
            fwhm_y.append(gaussian_sigma_to_fwhm * fit.y_stddev_1)
            print('Still fitting!!')
        else:
            sc = data_properties(cutout.data - sky)

            # dat = np.where(cutout.data - sky > 0, cutout.data - sky, 0)
            # mom1 = _moments(dat, order=1)
            # xc = mom1[0, 1] / mom1[0, 0]
            # yc = mom1[1, 0] / mom1[0, 0]
            # moments = _moments_central(dat,
            #                            center=(xc, yc), order=2)
            # mom_scale = (moments / mom1[0, 0])
            # fwhm_xm = 2 * np.sqrt(np.log(2) * mom_scale[0, 2])
            # fwhm_ym = 2 * np.sqrt(np.log(2) * mom_scale[2, 0])

            fwhm_xm = sc.fwhm.value
            fwhm_ym = fwhm_xm
            fwhm_x.append(fwhm_xm)
            fwhm_y.append(fwhm_ym)

    return np.array(fwhm_x), np.array(fwhm_y)


def source_detection(ccd, fwhm=8, sigma=3.0, iters=5,
                     threshold=10.0, find_fwhm=True,
                     sky_per_pix_avg=None):
    """
    Returns an astropy table containing the position of sources
    within the image.

    Parameters
    ----------

    ccd : `astropy.nddata.CCDData`
        The CCD Image array.

    fwhm : float, optional
        Full-width half-max of stars in the image.

    sigma : float, optional.
        The number of standard deviations to use as the lower and
        upper clipping limit.

    iters : int, optional
        The number of iterations to perform sigma clipping

    threshold : float, optional.
        The absolute image value above which to select sources.

    find_fwhm : bool, optional
        If ``True``, estimate the FWHM of each source by fitting a 2D Gaussian
        to it. If ``False``, estimate the FWHM of each source by computing
        the second moments of its light distribution.

    sky_per_pix_avg : float or array-like of float
        Sky background to subtract before centroiding.

    Returns
    -------

    sources: `astropy.table.Table`
        A table of the positions of sources in the image.  If `find_fwhm` is
        ``True``, includes a column called ``FWHM``.
    """
    mean, median, std = sigma_clipped_stats(ccd, sigma=sigma, maxiters=iters)
    daofind = DAOStarFinder(fwhm=fwhm, threshold=threshold * std)
    sources = daofind(ccd - median)
    print(sources)
    if find_fwhm:
        x, y = compute_fwhm(ccd, sources, fwhm_estimate=fwhm,
                            x_column='xcentroid', y_column='ycentroid',
                            sky_per_pix_avg=sky_per_pix_avg)
        sources['FWHM'] = (x + y) / 2
    return sources



from ..core import Camera


def test_camera_attributes():
    gain = 2.0
    read_noise = 10
    dark_current = 0.01
    c = Camera(gain=gain, read_noise=read_noise, dark_current=dark_current)
    assert c.gain == gain
    assert c.dark_current == dark_current
    assert c.read_noise == read_noise



import numpy as np

from ..magnitude_transforms import (filter_transform,
                                    calculate_transform_coefficients,
                                    transform_magnitudes)

import pytest

from astropy.modeling import models
from astropy.table import Table, Column
from astropy.utils.data import get_pkg_data_filename
from astropy import units as u


def generate_input_mags(n_stars):
    # Generate n_stars with magnitude in range 10 to 15
    rg = np.random.default_rng(1024)
    input_mags = rg.integers(0, high=50, size=n_stars) / 10 + 10
    instr_mags = Column(name='instrumental',
                        data=input_mags)
    return instr_mags


def generate_catalog_mags(instrument_mags, color, model):
    """
    Generate catalog magnitudes from instrumental magnitudes
    given a model that relates the two.
    """
    return instrument_mags + model(color)


def generate_star_coordinates(n_stars,
                              ra_start=180 * u.degree,
                              dec_start=45 * u.degree,
                              separation=10 * u.arcsec):
    """
    Generate RA/Dec coordinates for a set of stars.
    """
    # The plus one guarantees we'll have enough positions for
    # all of our stars.
    max_index = int(np.sqrt(n_stars)) + 1

    grids = np.mgrid[:max_index, :max_index]
    dec_grid, ra_grid = grids
    dec_offsets = separation * dec_grid
    ra_offsets = separation * ra_grid
    ra = (ra_start + ra_offsets).flatten()
    dec = (dec_start + dec_offsets).flatten()

    # Slice to return the correct number of positions.
    return ra[:n_stars], dec[:n_stars]


def generate_tables(n_stars, mag_model):
    """
    Generate both tables needed for transforming magnitudes.
    """
    instr_mags = generate_input_mags(n_stars)

    # Set name to match default value in function.
    instr_mags.name = 'mag_inst_r'

    # Set name to be default name for color.
    color = Column(name='B-V',
                   data=np.linspace(0.0, 1.0, num=len(instr_mags)))
    catalog = generate_catalog_mags(instr_mags, color, mag_model)

    # Again, set default name.
    catalog.name = 'r_mag'

    # We'll use the same RA/Dec for the catalog and and the instrumental
    # magnitudes.
    ra, dec = generate_star_coordinates(n_stars)

    # Instrumental magnitudes
    ra_col = Column(name='RA', data=ra)
    dec_col = Column(name='Dec', data=dec)

    instrumental = Table([instr_mags, ra_col, dec_col])

    # Yes, these really do need to be renamed for the catalog table
    ra_col.name = 'RAJ2000'
    dec_col.name = 'DEJ2000'

    catalog_table = Table([catalog, ra_col, dec_col, color])
    return instrumental, catalog_table


@pytest.mark.parametrize('bad_system', [None, 'monkeys'])
def test_filter_transform_bad_system(bad_system):
    fake_data = Table()
    with pytest.raises(ValueError) as e:
        filter_transform(fake_data, 'B', transform=bad_system)
    assert 'Must be one of' in str(e.value)
    assert str(bad_system) in str(e.value)


@pytest.mark.parametrize('system', ['ivezic', 'jester'])
def test_filter_transform(system):
    data_file = get_pkg_data_filename('data/mag_transform.csv')
    data = Table.read(data_file)
    in_system = data['system'] == system
    data = data[in_system]
    for output_filter in ['B', 'V', 'R', 'I']:
        f = filter_transform(data, output_filter, g='g', r='r', i='i',
                             transform=system)
        np.testing.assert_allclose(f, data[output_filter])


def test_filter_transform_bad_filter():
    with pytest.raises(ValueError) as e:
        filter_transform([], 'not a filter', transform='jester')
    assert 'the desired filter must be a string R B V or I' in str(e)


@pytest.mark.parametrize('order', [1, 2, 5])
def test_catalog_same_as_input(order):
    # Check that we get the correct transform when catalog magnitudes
    # are identical to instrument magnitudes.
    instr_mags = Column(name='instrumental', data=[10, 12.5, 11])
    zero = models.Const1D(0.0)
    color = Column(name='color', data=[1.0] * len(instr_mags))
    catalog = generate_catalog_mags(instr_mags, color, zero)

    _, fit_model = calculate_transform_coefficients(instr_mags,
                                                    catalog,
                                                    color,
                                                    order=order)
    assert len(fit_model.parameters) == order + 1
    assert all(fit_model.parameters == 0)


@pytest.mark.parametrize('order', [1, 2, 5])
def test_catalog_linear_to_input(order):
    # Check that we recover the correct relationship between
    # the catalog and instrumental magnitudes when the relationship
    # between the two is linear.
    n_stars = 100
    instr_mags = generate_input_mags(n_stars)
    true_relationship = models.Polynomial1D(1, c0=0.5, c1=0.75)
    color = Column(name='color',
                   data=np.linspace(0.0, 1.0, num=len(instr_mags)))
    catalog = generate_catalog_mags(instr_mags, color, true_relationship)

    _, fit_model = calculate_transform_coefficients(instr_mags,
                                                    catalog,
                                                    color,
                                                    order=order)
    assert len(fit_model.parameters) == order + 1
    assert np.abs(fit_model.c0 - true_relationship.c0) < 1e-7
    assert np.abs(fit_model.c1 - true_relationship.c1) < 1e-7

    if order >= 2:
        # Spot check some higher order terms -- they should be zero
        assert all(np.abs(fit_model.parameters[2:]) < 1e-7)


@pytest.mark.parametrize('order', [1, 2, 5])
def test_catalog_quadratic_to_input(order):
    # Check that we recover the correct relationship between
    # the catalog and instrumental magnitudes when the relationship
    # between the two is linear.
    n_stars = 100
    instr_mags = generate_input_mags(n_stars)
    true_relationship = models.Polynomial1D(2, c0=0.5, c1=0.75, c2=0.25)
    color = Column(name='color',
                   data=np.linspace(0.0, 1.0, num=len(instr_mags)))
    catalog = generate_catalog_mags(instr_mags, color, true_relationship)
    _, fit_model = calculate_transform_coefficients(instr_mags,
                                                    catalog,
                                                    color,
                                                    order=order)
    assert len(fit_model.parameters) == order + 1
    if order >= 2:
        # We expect a good fit in this case
        assert np.abs(fit_model.c0 - true_relationship.c0) < 1e-7
        assert np.abs(fit_model.c1 - true_relationship.c1) < 1e-7
        assert np.abs(fit_model.c2 - true_relationship.c2) < 1e-7
    else:
        # But a line just can't fit a quadratic that well
        assert np.abs(fit_model.c0 - true_relationship.c0) > 1e-7
        assert np.abs(fit_model.c1 - true_relationship.c1) > 1e-7
    if order >= 2:
        # Spot check some higher order terms -- they should be zero
        assert all(np.abs(fit_model.parameters[3:]) < 1e-7)


@pytest.mark.parametrize('faintest_magnitude', [None, 14])
def test_faintest_magnitude_has_effect(faintest_magnitude):
    # Check that the limit on magnitude when doing fits is respected.
    # We'll do this by setting up a linear relationship then
    # setting the catalog data fainter than the limit to nonsense.
    # Two outcomes we expect:
    #
    # 1. Without a limiting magnitude the fit should not be very good.
    # 2. With a limit the fit should be as good as it was before.
    n_stars = 100

    instr_mags = generate_input_mags(n_stars)
    true_relationship = models.Polynomial1D(1, c0=0.5, c1=0.75)
    color = Column(name='color',
                   data=np.linspace(0.0, 1.0, num=len(instr_mags)))
    catalog = generate_catalog_mags(instr_mags, color, true_relationship)

    faint_ones = catalog >= 14
    assert faint_ones.sum() > 0
    assert faint_ones.sum() < n_stars / 2

    # Scramble the faint ones
    rg = np.random.default_rng(40482)
    catalog[faint_ones] = (catalog[faint_ones]
                           + 5 * rg.random(faint_ones.sum()))

    _, fit_model = calculate_transform_coefficients(
                        instr_mags,
                        catalog,
                        color,
                        order=1,
                        faintest_mag=faintest_magnitude,
                        sigma=5000 # So that nothing is clipped
                    )

    if faintest_magnitude:
        assert np.abs(fit_model.c0 - true_relationship.c0) < 1e-7
        assert np.abs(fit_model.c1 - true_relationship.c1) < 1e-7
    else:
        assert np.abs(fit_model.c0 - true_relationship.c0) > 1e-2
        assert np.abs(fit_model.c1 - true_relationship.c1) > 1e-2


@pytest.mark.parametrize('order', [1, 2, 5])
def test_transform_magnitudes_identical_input(order):
    # Analogous to the test case for calculate_transform_coefficients
    # above where the input magnitudes are identical, except the input
    # objects have coordinates.
    n_stars = 100

    zero = models.Const1D(0.0)

    instrumental, catalog_table = generate_tables(n_stars, zero)

    calib_mags, stars_with_match, transform = \
        transform_magnitudes(instrumental, catalog_table, catalog_table,
                             order=order)

    print(calib_mags)
    assert all(calib_mags == catalog_table['r_mag'])
    assert all(stars_with_match)
    assert len(transform.parameters) == order + 1
    assert all(transform.parameters == 0)


@pytest.mark.parametrize('order', [1, 2, 5])
def test_transform_magnitudes_identical_coord_quad_mags(order):
    # Analogous to the test case for calculate_transform_coefficients
    # above where the input magnitudes are identical, except the input
    # objects have coordinates.
    n_stars = 100

    true_relationship = models.Polynomial1D(2, c0=0.5, c1=0.75, c2=0.25)

    instrumental, catalog_table = generate_tables(n_stars, true_relationship)

    calib_mags, stars_with_match, transform = \
        transform_magnitudes(instrumental, catalog_table, catalog_table,
                             order=order)

    assert all(stars_with_match)
    assert len(transform.parameters) == order + 1
    if order >= 2:
        # We expect a good fit in this case
        np.testing.assert_allclose(calib_mags, catalog_table['r_mag'],
                                   rtol=1e-7, atol=1e-7)
        assert np.abs(transform.c0 - true_relationship.c0) < 1e-7
        assert np.abs(transform.c1 - true_relationship.c1) < 1e-7
        assert np.abs(transform.c2 - true_relationship.c2) < 1e-7
    else:
        # But a line just can't fit a quadratic that well
        assert (np.abs(calib_mags - catalog_table['r_mag']) > 1e-5).all()
        assert np.abs(transform.c0 - true_relationship.c0) > 1e-7
        assert np.abs(transform.c1 - true_relationship.c1) > 1e-7
    if order >= 2:
        # Spot check some higher order terms -- they should be zero
        assert all(np.abs(transform.parameters[3:]) < 1e-7)


def test_coordinate_mismatches():
    # Test that stars without close coordinate matches end up
    # marked appropriately.
    n_stars = 100

    true_relationship = models.Polynomial1D(1, c0=0.5, c1=0.75)

    instrumental, catalog_table = generate_tables(n_stars, true_relationship)

    # Mess up the coordinates of half of the stars so that they don't match.
    catalog_table['RAJ2000'][50:] = (catalog_table['RAJ2000'][50:] +
                                     0.5 * u.degree)

    calib_mags, stars_with_match, transform = \
        transform_magnitudes(instrumental, catalog_table, catalog_table[:50],
                             order=2)

    assert all(stars_with_match[:50])
    assert all(~stars_with_match[50:])


def test_coordinate_all_mismatches():
    # Test that when no stars match stuff goes badly.
    n_stars = 100

    true_relationship = models.Polynomial1D(1, c0=0.5, c1=0.75)

    instrumental, catalog_table = generate_tables(n_stars, true_relationship)

    # Mess up the coordinates of half of the stars so that they don't match.
    catalog_table['RAJ2000'] = catalog_table['RAJ2000'] + 0.5 * u.degree

    calib_mags, stars_with_match, transform = \
        transform_magnitudes(instrumental, catalog_table, catalog_table[:50],
                             order=2)

    assert not any(stars_with_match)



import warnings

import numpy as np
from photutils.centroids import centroid_com
import ipywidgets as ipw

from astropy.io import fits
from astropy.stats import sigma_clipped_stats
from astropy.table import Table
from astropy.nddata import Cutout2D

try:
    from astrowidgets import ImageWidget
except ImportError:
    from astrowidgets.ginga import ImageWidget

import matplotlib.pyplot as plt

from stellarphot.io import TessSubmission
from stellarphot.visualization import seeing_plot
from stellarphot.visualization.fits_opener import FitsOpener

__all__ = ['set_keybindings', 'find_center', 'radial_profile',
            'find_hwhm', 'RadialProfile', 'box', 'SeeingProfileWidget']

desc_style = {"description_width": "initial"}


def set_keybindings(image_widget, scroll_zoom=False):
    """
    Set image widget keyboard bindings. The bindings are:

    + Pan by click-and-drag or with arrow keys.
    + Zoom by scrolling or using the ``+``/``-`` keys.
    + Adjust contrast by Ctrl-right click and drag
    + Reset contrast with shift-right-click.

    Any existing key bindings are removed.

    Parameters
    ----------

    image_widget : `astrowidgets.ImageWidget`
        Image widget on which to set the key bindings.

    scroll_zoom : bool, optional
        If True, zooming can be done by scrolling the mouse wheel.
        Default is False.

    Returns
    -------

    None
        Adds key bindings to the image widget.
    """
    bind_map = image_widget._viewer.get_bindmap()
    # Displays the event map...
    # bind_map.eventmap
    bind_map.clear_event_map()
    bind_map.map_event(None, (), 'ms_left', 'pan')
    if scroll_zoom:
        bind_map.map_event(None, (), 'pa_pan', 'zoom')

    # bind_map.map_event(None, (), 'ms_left', 'cursor')
    # contrast with right mouse
    bind_map.map_event(None, (), 'ms_right', 'contrast')

    # shift-right mouse to reset contrast
    bind_map.map_event(None, ('shift',), 'ms_right', 'contrast_restore')
    bind_map.map_event(None, ('ctrl',), 'ms_left', 'cursor')

    # Bind +/- to zoom in/out
    bind_map.map_event(None, (), 'kp_+', 'zoom_in')
    bind_map.map_event(None, (), 'kp_=', 'zoom_in')
    bind_map.map_event(None, (), 'kp_-', 'zoom_out')
    bind_map.map_event(None, (), 'kp__', 'zoom_out')

    # Bind arrow keys to panning
    # There is NOT a typo below. I want the keys to move the image in the
    # direction of the arrow
    bind_map.map_event(None, (), 'kp_left', 'pan_right')
    bind_map.map_event(None, (), 'kp_right', 'pan_left')
    bind_map.map_event(None, (), 'kp_up', 'pan_down')
    bind_map.map_event(None, (), 'kp_down', 'pan_up')


def find_center(image, center_guess, cutout_size=30, max_iters=10):
    """
    Find the centroid of a star from an initial guess of its position. Originally
    written to find star from a mouse click.

    Parameters
    ----------

    image : `astropy.nddata.CCDData` or numpy array
        Image containing the star.

    center_guess : array or tuple
        The position, in pixels, of the initial guess for the position of
        the star. The coordinates should be horizontal first, then vertical,
        i.e. opposite the usual Python convention for a numpy array.

    cutout_size : int, optional
        The default width of the cutout to use for finding the star.

    max_iters : int, optional
        Maximum number of iterations to go through in finding the center.

    Returns
    -------

    cen : array
        The position of the star, in pixels, as found by the centroiding
        algorithm.
    """
    pad = cutout_size // 2
    x, y = center_guess

    # Keep track of iterations
    cnt = 0

    # Grab the cutout...
    sub_data = image[y - pad:y + pad, x - pad:x + pad]  # - med

    # ...do stats on it...
    _, sub_med, _ = sigma_clipped_stats(sub_data)
    # sub_med = 0

    # ...and centroid.
    x_cm, y_cm = centroid_com(sub_data - sub_med)

    # Translate centroid back to original image (maybe use Cutout2D instead)
    cen = np.array([x_cm + x - pad, y_cm + y - pad])

    # ceno is the "original" center guess, set it to something nonsensical here
    ceno = np.array([-100, -100])

    while (cnt <= max_iters and
           (np.abs(np.array([x_cm, y_cm]) - pad).max() > 3
            or np.abs(cen - ceno).max() > 0.1)):

        # Update x, y positions for subsetting
        x = int(np.floor(x_cm)) + x - pad
        y = int(np.floor(y_cm)) + y - pad
        sub_data = image[y - pad:y + pad, x - pad:x + pad]  # - med
        _, sub_med, _ = sigma_clipped_stats(sub_data)
        # sub_med = 0
        mask = (sub_data - sub_med) < 0
        x_cm, y_cm = centroid_com(sub_data - sub_med, mask=mask)
        ceno = cen
        cen = np.array([x_cm + x - pad, y_cm + y - pad])
        if not np.all(~np.isnan(cen)):
            raise RuntimeError('Centroid finding failed, '
                               'previous was {}, current is {}'.format(ceno, cen))
        cnt += 1

    return cen


def radial_profile(data, center, size=30, return_scaled=True):
    """
    Construct a radial profile of a chunk of width ``size`` centered
    at ``center`` from image ``data`.

    Parameters
    ----------

    data : `astropy.nddata.CCDData` or numpy array
        Image data

    center : list-like
        x, y position of the center in pixel coordinates, i.e. horizontal
        coordinate then vertical.

    size : int, optional
        Width of the rectangular cutout to use in constructing the profile.

    return_scaled : bool, optional
        If ``True``, return an average radius and profile, otherwise
        it is cumulative. Not at all clear what a "cumulative" radius
        means, tbh.

    Returns
    -------

    r_exact : numpy array
        Exact radius of center of each pixels from profile center.

    ravg : numpy array
        Average radius in pixels used in constructing profile.

    radialprofile : numpy array
        Radial profile.
    """
    yd, xd = np.indices((size, size))

    sub_image = Cutout2D(data, center, size, mode='strict')
    sub_center = sub_image.center_cutout

    r = np.sqrt((xd - sub_center[0])**2 + (yd - sub_center[1])**2)
    r_exact = r.copy()
    r = r.astype(int)

    sub_data = sub_image.data

    tbin = np.bincount(r.ravel(), sub_data.ravel())
    rbin = np.bincount(r.ravel(), r_exact.ravel())
    nr = np.bincount(r.ravel())
    if return_scaled:
        radialprofile = tbin / nr
        ravg = rbin / nr
    else:
        radialprofile = tbin
        ravg = rbin

    return r_exact, ravg, radialprofile


def find_hwhm(r, intensity):
    """
    Estimate the half-width half-max from normalized, angle-averaged intensity profile.

    Parameters
    ----------

    r : array
        Radius of each pixel from the center of the star.

    intensity : array
        Normalized intensity at each radius.

    Returns
    -------

    r_half : float
        Radius at which the intensity is 50% the maximum
    """

    # Make the bold assumption that intensity decreases monotonically
    # so that we just need to find the first place where intensity is
    # less than 0.5 to estimate the HWHM.
    less_than_half = intensity < 0.5
    half_index = np.arange(len(less_than_half))[less_than_half][0]
    before_half = half_index - 1

    # Do linear interpolation to find the radius at which the intensity
    # is 0.5.
    r_more = r[before_half]
    r_less = r[half_index]
    I_more = intensity[before_half]
    I_less = intensity[half_index]

    I_half = 0.5

    r_half = r_less - (I_less - I_half) / (I_less - I_more) * (r_less - r_more)

    return r_half


class RadialProfile:
    """
    Class to hold radial profile information for a star.

    Parameters
    ----------

    data : numpy array
        Image data.

    x : int
        x position of the star.

    y : int
        y position of the star.

    Attributes
    ----------

    cen : tuple
        x, y position of the center of the star.

    data : numpy array
        Image data.

    FWHM : float
        Full-width half-max of the radial profile.

    profile_size : int
        Size of the cutout used to construct the radial profile.

    radius_values : numpy array
        Radius values for the radial profile.

    r_exact : numpy array
        Exact radius of center of each pixels from profile center.

    ravg : numpy array
        Average radius in pixels used in constructing profile.

    """
    def __init__(self, data, x, y):
        """
        Initialize the radial profile object instance.  Sets the center
        of the star and the image data.

        Parameters
        ----------

        data : numpy array
            Image data.

        x : int
            x position of the star.

        y : int
            y position of the star.


        """
        self.cen = find_center(data, (x, y), cutout_size=30)
        self.data = data

    def profile(self, profile_size):
        """
        Construct the radial profile of the star.  Sets
        ``r_exact``, ``ravg``, and ``radialprofile`` attributes.

        Parameters
        ----------

        profile_size : int
            Size of the cutout to use in constructing the profile.

        """
        self.profile_size = profile_size
        self.r_exact, self.ravg, self.radialprofile = (
            radial_profile(self.data,
                           self.cen,
                           size=profile_size)
        )

        self.sub_data = Cutout2D(self.data, self.cen, size=profile_size).data
        sub_med = np.median(self.sub_data)
        adjust_max = self.radialprofile.max() - sub_med
        self.scaled_profile = (self.radialprofile - sub_med) / adjust_max
        self.scaled_exact_counts = (self.sub_data - sub_med) / adjust_max
        self.HWHM = find_hwhm(self.ravg, self.scaled_profile)

    @property
    def FWHM(self):
        return int(np.round(2 * self.HWHM))

    @property
    def radius_values(self):
        return np.arange(len(self.radialprofile))


def box(imagewidget):
    """
    Compatibility layer for older versions of the photometry notebooks.

    Parameters
    ----------

    imagewidget : `astrowidgets.ImageWidget`
        ImageWidget instance to use for the seeing profile.

    Returns
    -------

    box : `ipywidgets.VBox`
        Box containing the seeing profile widget.
    """
    return SeeingProfileWidget(imagewidget=imagewidget).box


class SeeingProfileWidget:
    """
    A class for storing an instance of a widget displaying the seeing profile of stars in an image.

    Parameters
    ----------
    imagewidget : `astrowidgets.ImageWidget`, optional
        ImageWidget instance to use for the seeing profile.

    width : int, optional
        Width of the seeing profile widget.
    """
    def __init__(self, imagewidget=None, width=500):
        if not imagewidget:
            imagewidget = ImageWidget(image_width=width,
                                      image_height=width,
                                      use_opencv=True)

        self.iw = imagewidget
        # Do some set up of the ImageWidget
        set_keybindings(self.iw, scroll_zoom=False)
        bind_map = self.iw._viewer.get_bindmap()
        bind_map.map_event(None, ('shift',), 'ms_left', 'cursor')
        gvc = self.iw._viewer.get_canvas()
        self._mse = self._make_show_event()
        gvc.add_callback('cursor-down', self._mse)

        # Outputs to hold the graphs
        self.out = ipw.Output()
        self.out2 = ipw.Output()
        self.out3 = ipw.Output()
        # Build the larger widget
        self.container = ipw.VBox()
        self.fits_file = FitsOpener(title="Choose an image")
        big_box = ipw.HBox()
        big_box = ipw.GridspecLayout(1, 2)
        layout = ipw.Layout(width='20ch')
        hb = ipw.HBox()
        self.ap_t = ipw.IntText(description='Aperture radius', value=5, layout=layout, style=desc_style)
        self.in_t = ipw.IntText(description='Inner annulus', value=10, layout=layout, style=desc_style)
        self.out_t = ipw.IntText(description='Outer annulus', value=20, layout=layout, style=desc_style)
        self.save_aps = ipw.Button(description="Save settings")
        hb.children = [self.ap_t, self.save_aps] #, self.in_t, self.out_t]

        lil_box = ipw.VBox()
        lil_tabs = ipw.Tab()
        lil_tabs.children = [self.out3, self.out, self.out2]
        lil_tabs.set_title(0, "SNR")
        lil_tabs.set_title(1, "Seeing profile")
        lil_tabs.set_title(2, "Integrated counts")
        self.tess_box = self._make_tess_box()
        lil_box.children = [lil_tabs, self.tess_box]

        imbox = ipw.VBox()
        imbox.children = [imagewidget, hb]
        big_box[0, 0] = imbox
        big_box[0, 1] = lil_box
        big_box.layout.width = '100%'

        # Line below puts space between the image and the plots so the plots
        # don't jump around as the image value changes.
        big_box.layout.justify_content = 'space-between'
        self.big_box = big_box
        self.container.children = [self.fits_file.file_chooser, self.big_box]
        self.box = self.container
        self._aperture_name = 'aperture'

        self._tess_sub = None

        # Fill this in later with name of object from FITS file
        self.object_name = ''
        self._set_observers()

    def load_fits(self, file):
        self.fits_file.load_in_image_widget(self.iw)

    def _update_file(self, change):
        self.load_fits(change.selected)

    def _construct_tess_sub(self):
        file = self.fits_file.path
        self._tess_sub = TessSubmission.from_header(
            fits.getheader(file),
            telescope_code=self.setting_box.telescope_code.value,
            planet=self.setting_box.planet_num.value
        )

    def _set_seeing_profile_name(self, change):
        self._construct_tess_sub()
        self.seeing_file_name.value = self._tess_sub.seeing_profile

    def _save_toggle_action(self, change):
        activated = change['new']

        if activated:
            self.setting_box.layout.visibility = "visible"
            self._set_seeing_profile_name("")
        else:
            self.setting_box.layout.visibility = "hidden"

    def _save_seeing_plot(self, button):
        self._seeing_plot_fig.savefig(self.seeing_file_name.value)

    def _set_observers(self):
        def aperture_obs(change):
            self._mse(self.iw, aperture=change['new'])

        self.ap_t.observe(aperture_obs, names='value')
        self.save_aps.on_click(self._save_ap_settings)
        self.fits_file.register_callback(self._update_file)
        self.save_toggle.observe(self._save_toggle_action, names='value')
        self.save_seeing.on_click(self._save_seeing_plot)
        self.setting_box.planet_num.observe(self._set_seeing_profile_name)
        self.setting_box.telescope_code.observe(self._set_seeing_profile_name)

    def _save_ap_settings(self, button):
        ap_rad = self.ap_t.value
        with open('aperture_settings.txt', 'w') as f:
            f.write(f'{ap_rad},{ap_rad + 10},{ap_rad + 15}')

    def _make_tess_box(self):
        box = ipw.VBox()
        setting_box = ipw.HBox()
        self.save_toggle = ipw.ToggleButton(description="TESS seeing profile...",
                                            disabled=True)
        scope_name = ipw.Text(description="Telescope code",
                              value="Paul-P-Feder-0.4m",
                              style=desc_style)
        planet_num = ipw.IntText(description="Planet", value=1)
        self.save_seeing = ipw.Button(description="Save")
        self.seeing_file_name = ipw.Label(value="Moo")
        setting_box.children = (scope_name, planet_num, self.seeing_file_name, self.save_seeing)
        # for kid in setting_box.children:
        #     kid.disabled = True
        box.children = (self.save_toggle, setting_box)
        setting_box.telescope_code = scope_name
        setting_box.planet_num = planet_num
        setting_box.layout.flex_flow = "row wrap"
        setting_box.layout.visibility = "hidden"
        self.setting_box = setting_box
        return box

    def _make_show_event(self):

        def show_event(viewer, event=None, datax=None, datay=None, aperture=None):
            profile_size = 60
            fig_size = (10, 5)

            self.save_toggle.disabled = False

            if event is not None:
                # User clicked on a star, so generate profile
                i = self.iw._viewer.get_image()
                data = i.get_data()

                # Rough location of click in original image
                x = int(np.floor(event.data_x))
                y = int(np.floor(event.data_y))

                rad_prof = RadialProfile(data, x, y)

                try:
                    try: # Remove previous marker
                        self.iw.remove_markers(marker_name=self._aperture_name)
                    except AttributeError:
                        self.iw.remove_markers_by_name(marker_name=self._aperture_name)
                except ValueError:
                    # No markers yet, keep going
                    pass

                # ADD MARKER WHERE CLICKED
                self.iw.add_markers(Table(data=[[rad_prof.cen[0]], [rad_prof.cen[1]]],
                                          names=['x', 'y']),
                                    marker_name=self._aperture_name)

                # ----> MOVE PROFILE CONSTRUCTION INTO FUNCTION <----

                # CONSTRUCT RADIAL PROFILE OF PATCH AROUND STAR
                # NOTE: THIS IS NOT BACKGROUND SUBTRACTED
                rad_prof.profile(profile_size)

                # Default is 1.5 times FWHM
                aperture_radius = np.round(1.5 * 2 * rad_prof.HWHM, 0)
                self.rad_prof = rad_prof

                # Set this AFTER the radial profile has been created to avoid an attribute
                # error.
                self.ap_t.value = aperture_radius
            else:
                # User changed aperture
                aperture_radius = aperture
                rad_prof = self.rad_prof

            # DISPLAY THE SCALED PROFILE
            self.out.clear_output(wait=True)
            with self.out:
                # sub_med += med
                self._seeing_plot_fig = seeing_plot(rad_prof.r_exact, rad_prof.scaled_exact_counts,
                            rad_prof.ravg,
                            rad_prof.scaled_profile, rad_prof.HWHM,
                            self.object_name, gap=10, annulus_width=15,
                            radius = aperture_radius,
                            figsize=fig_size)
                plt.show()

            # CALCULATE AND DISPLAY NET COUNTS INSIDE RADIUS
            self.out2.clear_output(wait=True)
            with self.out2:
                sub_blot = rad_prof.sub_data.copy().astype('float32')
                min_idx = profile_size // 2 - 2 * rad_prof.FWHM
                max_idx = profile_size // 2 + 2 * rad_prof.FWHM
                sub_blot[min_idx:max_idx, min_idx:max_idx] = np.nan
                sub_std = np.nanstd(sub_blot)
                new_sub_med = np.nanmedian(sub_blot)
                r_exact, ravg, tbin2 = radial_profile(rad_prof.data - new_sub_med, rad_prof.cen,
                                                      size=profile_size,
                                                      return_scaled=False)
                r_exact_s, ravg_s, tbin2_s = radial_profile(rad_prof.data - new_sub_med, rad_prof.cen,
                                                      size=profile_size,
                                                      return_scaled=True)
                #tbin2 = np.bincount(r.ravel(), (sub_data - sub_med).ravel())
                counts = np.cumsum(tbin2)
                plt.figure(figsize=fig_size)
                plt.plot(rad_prof.radius_values, counts)
                plt.xlim(0, 40)
                ylim = plt.ylim()
                plt.vlines(aperture_radius, *plt.ylim(), colors=['red'])
                plt.ylim(*ylim)
                plt.grid()

                plt.title('Net counts in aperture')
                e_sky = np.nanmax([np.sqrt(new_sub_med), sub_std])

                plt.xlabel('Aperture radius (pixels)')
                plt.ylabel('Net counts')
                plt.show()

            # CALCULATE And DISPLAY SNR AS A FUNCTION OF RADIUS
            self.out3.clear_output(wait=True)
            with self.out3:
                read_noise = 10  # electrons
                gain = 1.5  # electrons/count
                # Poisson error is square root of the net number of counts enclosed
                poisson = np.sqrt(np.cumsum(tbin2))

                # This is obscure, but correctly calculated the number of pixels at
                # each radius, since the smoothed is tbin2 divided by the number of
                # pixels.
                nr = tbin2 / tbin2_s

                # This ignores dark current
                error = np.sqrt(poisson ** 2 + np.cumsum(nr)
                                * (e_sky ** 2 + (read_noise / gain)** 2))

                snr = np.cumsum(tbin2) / error
                plt.figure(figsize=fig_size)
                plt.plot(rad_prof.radius_values + 1, snr)

                plt.title(f'Signal to noise ratio max {snr.max():.1f} '
                          f'at radius {snr.argmax() + 1}')
                plt.xlim(0, 40)
                ylim = plt.ylim()
                plt.vlines(aperture_radius, *plt.ylim(), colors=['red'])
                plt.ylim(*ylim)
                plt.xlabel('Aperture radius (pixels)')
                plt.ylabel('SNR')
                plt.grid()
                plt.show()
        return show_event




# Licensed under a 3-clause BSD style license - see LICENSE.rst

from .multi_night_plots import *
from .aij_plots import *
from .comparison_functions import *
from .fits_opener import *
from .photometry_widget_functions import *
from .seeing_profile_functions import *
from .transit_plots import *



import sys

import pytest
import numpy as np

batman = pytest.importorskip('batman')

from stellarphot.analysis.transit_fitting import TransitModelFit


DEFAULT_TESTING_PARAMS = dict(
    t0=2, period=5, duration=0.1,
    depth=10, inclination=90,
    airmass_trend=0.005,
    width_trend=0.01,
    spp_trend=0.002
)

pytestmark = pytest.mark.skipif(sys.platform.startswith('win'),
                    reason="Intermittent hard crash on windows")


def _create_data_from_model_with_trends(transit_model, noise_dev=0.01):
    # Given a fully set up transit model, create some fake
    # data with a touch of noise.

    # Make our own batman model and data from it
    model = batman.TransitModel(transit_model._batman_params,
                                transit_model.times
                                )
    data = model.light_curve(transit_model._batman_params)

    if transit_model.airmass is not None:
        data += transit_model.model.airmass_trend * (transit_model.airmass)

    if transit_model.width is not None:
        data += transit_model.model.width_trend * transit_model.width

    if transit_model.spp is not None:
        data += transit_model.model.spp_trend * transit_model.spp

    if noise_dev > 0:
        # Make some noise
        generator = np.random.default_rng(432132)
        noise = generator.normal(scale=noise_dev, size=len(data))

        data += noise

    return data


def test_transit_fit_value_length_check():
    # Check that setting inconsistent lengths raises an error
    tmod = TransitModelFit()

    # Should work, all others are None
    tmod.times = np.array(list(range(5)))

    # Should work, new length is 5
    tmod.airmass = list(range(5))

    # Each of these should raise an error, has different length
    with pytest.raises(ValueError) as e:
        tmod.spp = list(range(3))
    assert 'Length of spp not consistent' in str(e.value)

    with pytest.raises(ValueError) as e:
        tmod.width = list(range(3))
    assert 'Length of width not consistent' in str(e.value)

    with pytest.raises(ValueError) as e:
        tmod.times = list(range(3))
    assert 'Length of times not consistent' in str(e.value)

    with pytest.raises(ValueError) as e:
        tmod.airmass = list(range(3))
    assert 'Length of airmass not consistent' in str(e.value)

    with pytest.raises(ValueError) as e:
        tmod.data = list(range(3))
    assert 'Length of data not consistent' in str(e.value)


@pytest.mark.parametrize('detrend_by', ['airmass', 'spp', 'width'])
def test_setting_unsetting_parameters_updates_detrend(detrend_by):
    tmod = TransitModelFit()

    dummy_values = np.array(list(range(5)))
    # Should work, all others are None
    tmod.times = dummy_values

    setattr(tmod, detrend_by, dummy_values)
    assert detrend_by in tmod._detrend_parameters

    setattr(tmod, detrend_by, None)
    assert detrend_by not in tmod._detrend_parameters


def test_transit_fit_setting_independent_vars():
    tmod = TransitModelFit()

    values = np.array(list(range(5)))

    for attr in ['times', 'airmass', 'width', 'spp']:
        setattr(tmod, attr, values)
        assert (getattr(tmod, attr) == values).all()


def test_transit_fit_setting_independent_vars_to_none():
    """
    Setting to None should always work.
    """
    tmod = TransitModelFit()

    values = np.array(list(range(5)))

    # First set values to something that is not None
    for attr in ['times', 'airmass', 'width', 'spp', 'data']:
        setattr(tmod, attr, values)

    # Now set to None and check value
    for attr in ['times', 'airmass', 'width', 'spp', 'data']:
        setattr(tmod, attr, None)
        assert getattr(tmod, attr) is None


def test_transit_create_model():
    # Creating a model should not give us an error and should lead
    # to expected values of parameters.
    tmod = TransitModelFit()

    tmod.setup_model(**DEFAULT_TESTING_PARAMS)

    duration = DEFAULT_TESTING_PARAMS['duration']
    period = DEFAULT_TESTING_PARAMS['period']

    expected_a = 1 / np.sin(duration * np.pi / period)
    assert tmod.model.t0 == DEFAULT_TESTING_PARAMS['t0']
    assert tmod.model.rp == np.sqrt(DEFAULT_TESTING_PARAMS['depth'] / 1000)
    assert np.abs(tmod.model.a - expected_a) < 1e-7
    assert tmod.model.period == DEFAULT_TESTING_PARAMS['period']
    assert tmod.model.airmass_trend == DEFAULT_TESTING_PARAMS['airmass_trend']
    assert tmod.model.width_trend == DEFAULT_TESTING_PARAMS['width_trend']
    assert tmod.model.spp_trend == DEFAULT_TESTING_PARAMS['spp_trend']


def _make_transit_model_with_data(noise_dev=1e-1,
                                  with_airmass=False,
                                  with_width=False,
                                  with_spp=False):
    tmod = TransitModelFit()

    tmod.setup_model(**DEFAULT_TESTING_PARAMS)

    t0 = DEFAULT_TESTING_PARAMS['t0']
    duration = DEFAULT_TESTING_PARAMS['duration']

    # Last factor ensures some out of transit data
    start = t0 - duration / 2 * 1.3
    end = t0 + duration / 2 * 1.3

    times = np.linspace(start, end, num=100)
    tmod.times = times

    # Make an airmass that starts at 1.4 and decreases to 1.0
    # over the duration and is parabolic centered on the final time.
    airmass = 0.4 / (times[0] - times[-1])**2 * (times - times[-1]) ** 2 + 1.0

    tmod.airmass = airmass if with_airmass else None

    # Make a width that increases linearly with time. Never happen IRL, but
    # that is ok. Let's start at 5 pixels and increase to 9.
    width = 4 * (times - times[0]) / (times[-1] - times[0]) + 5

    tmod.width = width if with_width else None

    # sky_per_pixel...let's make that sinusoidal. Not realistic but should be
    # easy for the fitter to pick out this trend.
    spp = 30 + 5 * np.sin(4 * np.pi * times / (times[-1] - times[0]))

    tmod.spp = spp if with_spp else None

    data = _create_data_from_model_with_trends(tmod, noise_dev=noise_dev)

    tmod.data = data

    return tmod


def test_transit_fit_all_parameters():
    tmod = _make_transit_model_with_data(noise_dev=1e-5,
                                         with_airmass=True,
                                         with_width=True,
                                         with_spp=True)

    duration = DEFAULT_TESTING_PARAMS['duration']
    period = DEFAULT_TESTING_PARAMS['period']

    expected_a = 1 / np.sin(duration * np.pi / period)

    assert np.abs(tmod.model.a - expected_a) < 1e-6

    tmod.fit()

    # Check the non-exoplanet trends
    for fit_trend in ['airmass_trend', 'width_trend', 'spp_trend']:
        assert (np.abs(getattr(tmod.model, fit_trend)
                       - DEFAULT_TESTING_PARAMS[fit_trend]) < 1e-5)

    # Check a few of the exoplanet parameters
    assert np.abs(tmod.model.t0 - DEFAULT_TESTING_PARAMS['t0']) < 1e-3

    assert np.abs(tmod.model.a - expected_a) < 1e-2

    expected_rp = np.sqrt(DEFAULT_TESTING_PARAMS['depth'] / 1000)
    assert np.abs(tmod.model.rp - expected_rp) < 1e-4

    assert 'airmass' in tmod._detrend_parameters
    assert 'width' in tmod._detrend_parameters
    assert 'spp' in tmod._detrend_parameters


def test_transit_model_detrend():
    tmod = _make_transit_model_with_data(noise_dev=0,
                                         with_airmass=True,
                                         with_width=True,
                                         with_spp=True)

    no_trends = _make_transit_model_with_data(noise_dev=0,
                                              with_airmass=False,
                                              with_width=False,
                                              with_spp=False)
    full_model = tmod.model_light_curve()

    np.testing.assert_allclose(full_model - no_trends.model_light_curve(),
                               tmod.model.airmass_trend * tmod.airmass +
                               tmod.model.width_trend * tmod.width +
                               tmod.model.spp_trend * tmod.spp)

    for trend in tmod._all_detrend_params:
        detrended_model = tmod.model_light_curve(detrend_by=trend)
        trend_param = getattr(tmod.model, f'{trend}_trend')
        trend_data = getattr(tmod, trend)
        assert trend_param != 0
        assert trend_data is not None
        np.testing.assert_allclose(full_model - detrended_model,
                                   trend_param * trend_data)

    detrended_model = tmod.model_light_curve(detrend_by='all')
    np.testing.assert_allclose(detrended_model, no_trends.model_light_curve())


def test_transit_data_detrend():
    tmod = _make_transit_model_with_data(noise_dev=0,
                                         with_airmass=True,
                                         with_width=True,
                                         with_spp=True)

    no_trends = _make_transit_model_with_data(noise_dev=0,
                                              with_airmass=False,
                                              with_width=False,
                                              with_spp=False)

    np.testing.assert_allclose(tmod.data_light_curve(detrend_by='all'),
                               no_trends.data)
    np.testing.assert_allclose(tmod.data - no_trends.data,
                               tmod.model.airmass_trend * tmod.airmass +
                               tmod.model.width_trend * tmod.width +
                               tmod.model.spp_trend * tmod.spp)


def test_transit_fit_parameters_unfreeze_as_expected():
    tmod = _make_transit_model_with_data(noise_dev=1e-5,
                                         with_airmass=False,
                                         with_width=False,
                                         with_spp=False)

    # None of these are fixed by default
    assert not tmod.model.airmass_trend.fixed
    assert not tmod.model.width_trend.fixed
    assert not tmod.model.spp_trend.fixed

    tmod.fit()

    # They should return to their original state after the
    # fit even though they are temporarily fixed during the
    # fit.
    assert not tmod.model.airmass_trend.fixed
    assert not tmod.model.width_trend.fixed
    assert not tmod.model.spp_trend.fixed



import numpy as np

from astropy.table import Table
from astropy.coordinates import SkyCoord
from astropy import units as u

__all__ = ['calc_multi_vmag', 'calc_vmag']


def calc_multi_vmag(var_stars, star_data, comp_stars, **kwd):
    """
    Calculate the average magnitude and standard deviation of multiple variable
    stars in a field.

    Parameters
    ----------
    var_stars : '~astropy.table.Table'
        Table of variable stars known in field

    star_data : '~astropy.table.Table'
        Table of star data from observation image

    comp_stars : '~astropy.table.Table'
        Table of known comparison stars in the field, given by AAVSO

    kwd : Keyword arguments passed through to calc_vmag

    Returns
    -------
    vmag_table : '~astropy.table.Table'
        Table including the name, avgerage magnitude, and standard deviation of
        multiple variable stars in a field.
    """
    name = []
    vmag = []
    stdev = []
    for vsx in var_stars:
        name.append(vsx['Name'])
        avg_vmag, error = calc_vmag(vsx, star_data, comp_stars, **kwd)
        vmag.append(avg_vmag)
        stdev.append(error)
    vmag_table = Table([name, vmag, stdev], names=('Name', 'Mag', 'StDev'))
    return vmag_table


def calc_vmag(var_stars, star_data, comp_stars, band=None,
              star_data_mag_column='mag_inst'):
    """
    Calculate the average magnitude and standard deviation of a variable star in field.

    Parameters
    ----------

    var_stars : '~astropy.table.Table'
        Table of variable stars known in field. It should contain a column
        called ``coords`` with the coordinates for each variable star as
        astropy ``SkyCoord`` objects.

    star_data : '~astropy.table.Table'
        Table of star data from observed image(s). One column should be named
        ``band`` and contain the passband in which observations were done.
        The column containing instrumental magnitudes is passed in with the
        argument ``star_data_mag_column``.

    comp_stars : '~astropy.table.Table'
        Table of known comparison stars in the field, given by AAVSO. The
        column containing the reference magnitudes for the filter specified
        by filter is passed in with the argument ``comp_star_mag_column``.

    band : str
        Filter/passband in which magnitude is being calculated.

    star_data_mag_column : str, optional
        Name of the column containing the instrumental magnitudes
        in ``star_data``.

    Returns
    -------
    avg : float
        Average magnitude for the variable star

    stdev : float
        Standard deviation for variable star values
    """

    if not band:
        raise ValueError("You must provide a band for this function.")

    # Match variable stars (essentially a list of targets) to instrumental
    # magnitude information.
    var_coords = var_stars['coords']
    star_data_coords = SkyCoord(ra=star_data['RA'], dec=star_data['Dec'])
    v_index, v_d2d, _ = var_coords.match_to_catalog_sky(star_data_coords)

    rcomps = comp_stars[comp_stars['band'] == band]

    # Match comparison star list to instrumental magnitude information
    try:
        comp_coords = rcomps['coords']
    except KeyError:
        comp_coords = SkyCoord(ra=rcomps['RAJ2000'], dec=rcomps['DEJ2000'])
    index, d2d, _ = comp_coords.match_to_catalog_sky(star_data_coords)
    good = d2d < 1 * u.arcsec
    good_index = index[good]

    vmag_image = star_data[v_index][star_data_mag_column]
    comp_star_mag = star_data[good_index][star_data_mag_column]
    a_index, a_d2d, _ = comp_coords.match_to_catalog_sky(comp_coords)
    good_a_index = a_index[good]
    accepted_comp = rcomps[good_a_index]['mag']
    new_mag = vmag_image - comp_star_mag + accepted_comp
    avg = np.nanmean(new_mag)
    stdev = np.nanstd(new_mag)

    return avg, stdev



from dataclasses import dataclass, field
from pathlib import Path
import re
from tempfile import NamedTemporaryFile

from astropy.coordinates import SkyCoord
from astropy.table import Table
from astropy.time import Time
from astropy import units as u
from astropy.utils.data import download_file

from astroquery.mast import Catalogs

import requests

from stellarphot.analysis.exotic import get_tic_info

__all__ = ["TessSubmission", "TOI", "TessTargetFile"]

# Makes me want to vomit, but....
DEFAULT_TABLE_LOCATION = "who.the.heck.knows"
TOI_TABLE_URL = "https://exofop.ipac.caltech.edu/tess/download_toi.php?output=csv"
TIC_regex = re.compile(r"[tT][iI][cC][^\d]?(?P<star>\d+)(?P<planet>\.\d\d)?")


@dataclass
class TessSubmission:
    """
    A data class to represent TESS submissions.

    Parameters
    ----------

    telescope_code: String
        The telescope code, e.g. "SRO" or "TJO"

    filter: String
        The filter used for the observations, e.g. "Ic" or "Rc"

    utc_start: String
        The UTC date of the first observation, in YYYYMMDD format

    tic_id: int
        The TIC ID of the target

    planet_number: int
        The planet number, if applicable


    Attributes
    ----------
    base_name: str
        The base name of the submission, e.g. "TIC123456789-01_20200101_SRO_Ic"

    seeing_profile: str
        The name of the seeing profile file, e.g. "TIC123456789-01_20200101_SRO_Ic_seeing-profile.png"

    field_image: str
        The name of the field image file, e.g. "TIC123456789-01_20200101_SRO_Ic_field.png"

    field_image_zoom: str
        The name of the zoomed-in field image file, e.g. "TIC123456789-01_20200101_SRO_Ic_field-zoom.png"

    apertures: str
        The name of the apertures file, e.g. "TIC123456789-01_20200101_SRO_Ic_measurements.apertures"

    tic_coord: `astropy.coordinates.SkyCoord`
        The SkyCoord of the target, from the TIC catalog.
    """
    telescope_code: str
    filter: str
    utc_start: int
    tic_id: int
    planet_number: int

    def __post_init__(self, *args, **kwargs):
        self._tic_info = None

    @classmethod
    def from_header(cls, header, telescope_code="", planet=0):
        """
        Create a TessSubmission from a FITS header

        Parameters
        ----------
        header: `astropy.io.fits.Header`
            The FITS header to parse

        telescope_code: str
            The telescope code, e.g. "SRO" or "TJO"

        planet: int
            The planet number, if applicable
        """

        tic_id = 0
        filter = ""
        fails = {}
        try:
            dateobs = header['date-obs']
        except KeyError:
            fails["utc_start"] = "UTC date of first image"
        else:
            dateobs = dateobs.split("T")[0].replace("-", "")

        try:
            filter = header['filter']
        except KeyError:
            fails["filter"] = ("filter/passband")

        try:
            obj = header['object']
        except KeyError:
            fails['tic_id'] = "TIC ID number"
        else:
            result = TIC_regex.match(obj)
            if result:
                tic_id = int(result.group("star"))
                # Explicit argument overrules the header
                if result.group("planet") and not planet:
                    # Drop the leading period from the match
                    planet = int(result.group("planet")[1:])
            else:
                # No star from the object after all
                fails['tic_id'] = "TIC ID number"

        fail_msg = ""
        fail = []
        for k, v in fails.items():
            fail.append(f"Unable to determine {k}, {v}, from header.")

        fail = "\n".join(fail)

        if fail:
            raise ValueError(fail)

        return cls(utc_start=dateobs,
                   filter=filter,
                   telescope_code=telescope_code,
                   tic_id=tic_id,
                   planet_number=planet)

    def _valid_tele_code(self):
        return len(self.telescope_code) > 0

    def _valid_planet(self):
        return self.planet_number > 0

    def _valid_tic_num(self):
        return self.tic_id < 10_000_000_000

    def _valid(self):
        """
        Check whether the information so far is valid, meaning:
         + Telescope code is not the empty string
         + Planet number is not zero
         + TIC ID is not more than 10 digits
        """
        valid = (
            self._valid_tele_code() and
            self._valid_planet() and
            self._valid_tic_num()
        )
        return valid

    @property
    def base_name(self):
        if self._valid():
            pieces = [
                f"TIC{self.tic_id}-{self.planet_number:02d}",
                self.utc_start,
                self.telescope_code,
                self.filter
            ]
            return "_".join(pieces)

    @property
    def seeing_profile(self):
        return self.base_name + "_seeing-profile.png"

    @property
    def field_image(self):
        return self.base_name + "_field.png"

    @property
    def field_image_zoom(self):
        return self.base_name + "_field-zoom.png"

    @property
    def apertures(self):
        return self.base_name + "_measurements.apertures"

    @property
    def tic_coord(self):
        if not self._tic_info:
            self._tic_info = get_tic_info(self.tic_id)
        return SkyCoord(ra=self._tic_info['ra'][0], dec=self._tic_info['dec'][0], unit='degree')

    def invalid_parts(self):
        if self._valid():
            return

        if not self._valid_tele_code():
            print(f"Telescope code {self.telescope_code} is not valid")

        if not self._valid_planet():
            print(f"Planet number {self.planet_number} is not valid")

        if not self._valid_tic_num():
            print(f"TIC ID {self.tic_id} is not valid.")


class TOI:
    """ A class to hold information about a TOI (TESS Object of Interest).

    Attributes
    ----------

    coord: `astropy.coordinates.SkyCoord`
        The coordinates of the target.

    depth: float
        The transit depth of the target.

    depth_error: float
        The uncertainty in the transit depth.

    duration: float
        The duration of the transit.

    duration_error: float
        The uncertainty in the duration of the transit.

    epoch: float
        The epoch of the transit.

    epoch_error: float
        The uncertainty in the epoch of the transit.

    period: float
        The period of the transit.

    period_error: float
        The uncertainty in the period of the transit.

    tess_mag: float
        The TESS magnitude of the target.

    tess_mag_error: float
        The uncertainty in the TESS magnitude.

    tic_id: int
        The TIC ID of the target.
    """
    def __init__(self, tic_id, toi_table=DEFAULT_TABLE_LOCATION, allow_download=True):
        """
        Initialize a TOI object.

        Parameters
        ----------

        tic_id: int
            The TIC ID of the target.

        toi_table: str, optional
            The path to the TOI table. If not provided, the default table will be downloaded.

        allow_download: bool, optional
            Whether to allow the default table to be downloaded if it is not found.

        """
        path = Path(toi_table)
        if not path.is_file():
            if not allow_download:
                raise ValueError(f"File {toi_table} not found.")
            toi_table = download_file(TOI_TABLE_URL, cache=True, show_progress=True, timeout=60)

        self._toi_table = Table.read(toi_table, format="ascii.csv")
        self._toi_table = self._toi_table[self._toi_table['TIC ID'] == tic_id]
        if len(self._toi_table) != 1:
            raise RuntimeError(f"Found {len(self._toi_table)} rows in table, expected one.")
        self._tic_info = get_tic_info(tic_id)

    @property
    def tess_mag(self):
        return self._toi_table['TESS Mag'][0]

    @property
    def tess_mag_error(self):
        return self._toi_table['TESS Mag err'][0]

    @property
    def depth(self):
        """
        Depth, parts per thousand.
        """
        return self._toi_table['Depth (ppm)'][0] / 1000

    @property
    def depth_error(self):
        """
        Error in depth, parts per thousand.
        """
        return self._toi_table['Depth (ppm) err'][0] / 1000

    @property
    def epoch(self):
        return Time(self._toi_table['Epoch (BJD)'][0], scale='tdb', format='jd')

    @property
    def epoch_error(self):
        return self._toi_table['Epoch (BJD) err'][0] * u.day

    @property
    def period(self):
        return self._toi_table['Period (days)'][0] * u.day

    @property
    def period_error(self):
        return self._toi_table['Period (days) err'][0] * u.day

    @property
    def duration(self):
        return self._toi_table['Duration (hours)'][0] * u.hour

    @property
    def duration_error(self):
        return self._toi_table['Duration (hours) err'][0] * u.hour

    @property
    def coord(self):
        return SkyCoord(ra=self._tic_info['ra'][0], dec=self._tic_info['dec'][0], unit='degree')

    @property
    def tic_id(self):
        return self._tic_info['ID'][0]

@dataclass
class TessTargetFile:
    """
    A class to hold information about a TESS target file.

    Parameters
    ----------

    coord : `astropy.coordinates.SkyCoord`
        The coordinates of the target.

    magnitude : float
        The magnitude of the target.

    depth : float
        The depth of the transit.

    file : str, optional
        The path to the target file. If not provided, a temporary file will be created.

    aperture_server : str, optional
        The URL of the aperture server. default: https://www.astro.louisville.edu/


    Attributes
    ----------

    aperture_server : str
        The URL of the aperture server.

    coord : `astropy.coordinates.SkyCoord`
        The coordinates of the target.

    depth : float
        The depth of the transit.

    file : str
        The path to the target file. If not provided, a temporary file will be created.

    magnitude : float
        The magnitude of the target.

    target_file : str
        The path to the target file.

    target_table : `astropy.table.Table`
        The target table.

    """
    coord : SkyCoord
    magnitude : float
    depth : float
    file : str = ""
    aperture_server : str = "https://www.astro.louisville.edu/"

    def __post_init__(self):
        if not self.file:
            self.file = NamedTemporaryFile()
        self._path = Path(self.file.name)
        self.target_file = self._retrieve_target_file()
        self.target_table = self._build_table()

    def _retrieve_target_file(self):
        params = dict(
            ra = self.coord.ra.to_string(unit='hour', decimal=False, sep=":"),
            dec = self.coord.dec.to_string(unit='degree', decimal=False, sep=":"),
            mag=self.magnitude,
            depth=self.depth
        )
        result = requests.get(self.aperture_server + "cgi-bin/gaia_to_aij/upload_request.cgi", params=params)
        links = re.search('href="(.+)"', result.text.replace('\n', ''), )
        download_link = self.aperture_server + links[1]
        target_file_contents = requests.get(download_link)
        with open(self._path, "w") as f:
            f.write(target_file_contents.text)


    def _build_table(self):
        from stellarphot.visualization.comparison_functions import read_file

        self.table = read_file(self._path)



import numpy as np


__all__ = ['plot_many_factors', 'bin_data', 'scale_and_shift']


def plot_many_factors(photometry, low, high, shift, scale):
    """ Plots many factors of photometry against each other.

    Parameters
    ----------

    photometry : `astropy.table.Table`
        The photometry table to plot.

    low : float
        The lower bound of the y-axis.

    high : float
        The upper bound of the y-axis.

    shift : float
        The amount to shift the data by.

    scale : float
        The amount to scale the data by.

    Returns
    -------

    None
        Added features to the plot directly.
    """
    airmass = photometry['airmass'] / np.mean(photometry['airmass'])
    x = photometry['xcenter'] / np.mean(photometry['xcenter'])
    y = photometry['ycenter'] / np.mean(photometry['ycenter'])
    comp_counts = photometry['comparison counts'] / np.mean(photometry['comparison counts'])
    sky_per_pix = photometry['sky_per_pix_avg'] / np.mean(photometry['sky_per_pix_avg'])
    width = photometry['width'] / np.mean(photometry['width'])

    scale_airmass = scale_and_shift(airmass, scale, 0.75 * shift, pos=False)
    scale_x = scale_and_shift(x, scale, shift, pos=True)
    scale_y = scale_and_shift(y, scale, shift, pos=True)
    scale_sky_pix = scale_and_shift(sky_per_pix, scale, shift, pos=True)
    scale_counts = scale_and_shift(comp_counts, scale, shift, pos=True)
    scale_width = scale_and_shift(width, scale, shift, pos=True)

    grid_y_ticks = np.arange(low, high, 0.02)


def bin_data(data_set, num=3, error_set=None):
    """ Bins data into groups of num.

    Parameters
    ----------

    data_set : array-like
        The data to bin.

    num : int, optional
        The number of data points to bin together. Default is 3.

    error_set : array-like
        The error on the data set. Default is None.

    Returns
    -------

    binned_set : array-like
        The binned data set.

    error : array-like, optional
        The error on the binned data set.
    """
    binned_set = []
    error = []
    for i in range(0, len(data_set), num):
        binned_set.append(data_set[i:i+num].mean())
        if error_set is not None:
            error_bin = error_set[i:i+num]**2
            error.append(error_bin.sum()/num)
    return np.array(binned_set), np.array(error)


def scale_and_shift(data_set, scale, shift, pos=True):
    """ Scales and shifts data set passed in.

    Parameters
    ----------

    data_set : array-like
        The data to scale and shift.

    scale : float
        The amount to scale the data by.

    shift : float
        The amount to shift the data by (in scaled units).

    pos : bool, optional
        Is data displayed in positive or negative direction? Default is True.

    Returns
    -------

    data_set : array-like
        The scaled and shifted data.
    """
    if not pos:
        data_set = 1 - scale * (data_set - data_set.min()) / (data_set.max() - data_set.min())
    else:
        data_set = 1 + scale * (data_set - data_set.min()) / (data_set.max() - data_set.min())

    data_set += shift

    return data_set



import bottleneck as bn
import numpy as np
from photutils import (DAOStarFinder, aperture_photometry, CircularAperture,
                       CircularAnnulus, extract_stars)
from photutils.centroids import centroid_sources

from astropy.coordinates import SkyCoord, EarthLocation
from astropy.table import vstack
from astropy import units as u
from astropy.time import Time

from ccdproc import ImageFileCollection

from astropy.stats import sigma_clipped_stats
from astropy.nddata import NoOverlapError

from .source_detection import compute_fwhm

__all__ = ['photutils_stellar_photometry',
           'faster_sigma_clip_stats',
           'find_too_close', 'clipped_sky_per_pix_stats',
           'add_to_photometry_table', 'photometry_on_directory',
           'calculate_noise', 'find_times']


def photutils_stellar_photometry(ccd_image, sources,
                                 aperture_radius, inner_annulus,
                                 outer_annulus, gain=1.0, N_R=0, N_dark_pp=0,
                                 reject_background_outliers=True):
    """
    Perform aperture photometry on an image, with a few options for estimating
    the local background from an annulus around the aperture.

    Parameters
    ----------
    ccd_image : `ccdproc.CCDData`
        Image on which to perform aperture photometry.

    sources : `astropy.table.Table`
        Table of extracted sources. Assumed to be the output of
        `photutils.daofind()` source extraction function.

    aperture_radius : float
        Radius of aperture(s) in pixels.

    inner_annulus : float
        Inner radius of the annulus in pixels.

    outer_annulus : float
        Outer radius of the annulus in pixels.

    gain : float
        Gain of the CCD. In units of electrons per DN.

    N_R : float
        Read noise of the CCD in electrons per pixel.

    N_dark_pp : float
        Number of dark counts per pixel.

    reject_background_outliers : bool, optional
        If ``True``, sigma clip the pixels in the annulus to reject outlying
        pixels (e.g. like stars in the annulus)

    Returns
    -------
    phot_table : `astropy.table.Table`
        Astropy table with columns for flux, x/y coordinates of center,
        RA/dec coordinates of center, sky background per pixel,
        net flux, aperture and annulus radii used, and flux error.
    """

    # check that the outer radius is greater or equal the inner radius
    # for annulus
    if inner_annulus >= outer_annulus:
        raise ValueError("outer_annulus must be greater than inner_annulus")

    # check that the annulus inner radius is greater or equal
    # the aperture radius
    if aperture_radius >= inner_annulus:
        raise ValueError("inner_radius must be greater than aperture_radius")

    # Extract x,y coordinates from sources table, construct aperture and
    # annulus objects from coordinates, and perform aperture photometry
    coords = np.array([sources['xcentroid'], sources['ycentroid']]).T
    apertures = CircularAperture(coords, aperture_radius)
    annulus = CircularAnnulus(coords, inner_annulus, outer_annulus)
    phot_table = aperture_photometry(ccd_image, apertures)
    phot_table_1 = aperture_photometry(ccd_image, annulus)

    # Obtain the local background/pixel and net flux between the aperture and
    # annulus objects
    n_pix_ap = apertures.area
    n_pix_ann = annulus.area

    if reject_background_outliers:
        annulus_masks = annulus.to_mask()
        bkgd_pp = []
        for mask in annulus_masks:
            annulus_data = mask.cutout(ccd_image)
            bool_mask = mask.data < 1
            # Only include whole pixels in the estimate of the background
            bkgd, _, _ = sigma_clipped_stats(annulus_data * mask,
                                             mask=bool_mask)
            bkgd_pp.append(bkgd)
        bkgd_pp = np.array(bkgd_pp)
    else:
        bkgd_pp = phot_table_1['aperture_sum'] / n_pix_ann

    net_flux = phot_table['aperture_sum'] - (n_pix_ap * bkgd_pp)
    phot_table['background_per_pixel'] = bkgd_pp
    phot_table['net_flux'] = net_flux

    # Return a columns with the aperture radius and
    # the inner/outer annulus radii
    phot_table['aperture_radius'] = \
        np.ones(len(phot_table['aperture_sum'])) * aperture_radius
    phot_table['inner_radius'] = \
        np.ones(len(phot_table['aperture_sum'])) * inner_annulus
    phot_table['outer_radius'] = \
        np.ones(len(phot_table['aperture_sum'])) * outer_annulus

    # Obtain RA/Dec coordinates and add them to table
    try:
        ra, dec = ccd_image.wcs.all_pix2world(coords[0], coords[1], 0)
        phot_table['RA_center'] = ra
        phot_table['Dec_center'] = dec
    except AttributeError:
        pass

    # Obtain flux error and add column to return table
    noise = np.sqrt(gain * net_flux + n_pix_ap * (1 + (n_pix_ap / n_pix_ann)) *
                    (gain * (bkgd_pp + N_dark_pp) + (N_R**2)))
    phot_table['aperture_sum_err'] = noise

    return phot_table


def faster_sigma_clip_stats(data, sigma=5, iters=5, axis=None):
    """
    Calculate sigma clipped stats quickly using NaNs instead of masking
    and using bottleneck where possible.

    Parameters
    ----------

    data : numpy array
        The data to be clipped. *The data should have had masked values
        replaced with NaN prior to calling this function.*

    sigma : float, optional
        Number of standard deviations (estimated with the MAD) a point must
        be from the central value (median) to be rejected.

    iters : int, optional
        Number of sigma clipping iterations to perform. Fewer iterations than
        this may be performed because iterations stop when no new data is
        being clipped.

    axis : int, optional
        axis along which to perform the median.

    Returns
    -------
    mean, median, std : float or numpy array
        Clipped statistics; shape depends on the shape of the input.
    """
    data = data.copy()
    for _ in range(iters):
        central = bn.nanmedian(data, axis=axis)
        try:
            central = central[:, np.newaxis]
        except (ValueError, IndexError, TypeError):
            pass

        std_dif = 1.4826 * bn.nanmedian(np.abs(data - central))

        clips = np.abs(data - central) / std_dif > sigma

        if np.nansum(clips) == 0:
            break
        data[clips] = np.nan
    return (bn.nanmean(data, axis=axis), bn.nanmedian(data, axis=axis),
            bn.nanstd(data, axis=axis))


def find_too_close(star_locs, aperture_rad, pixel_scale=0.563):
    """
    Identify sources that are closer together than twice the aperture radius.

    Parameters
    ----------

    star_locs : tuple of numpy array
        The first entry in the tuple should be the right ascension of the
        sources, in degrees. The second should be the declination of
        the sources, in degrees.

    aperture_rad : int
        Radius of the aperture, in pixels.

    pixel_scale : float, optional
        Pixel scale of the image in arcsec/pixel

    Returns
    -------

    numpy array of bool
        Array the same length as the RA/Dec that is ``True`` where the sources
        are closer than two aperture radii, ``False`` otherwise.
    """
    star_coords = SkyCoord(ra=star_locs[0], dec=star_locs[1],
                           frame='icrs', unit='degree')
    idxc, d2d, d3d = star_coords.match_to_catalog_sky(star_coords,
                                                      nthneighbor=2)
    return d2d < (aperture_rad * 2 * 0.563 * u.arcsec)


def clipped_sky_per_pix_stats(data, annulus, sigma=5, iters=5):
    """
    Calculate sigma-clipped statistics on an annulus.

    Parameters
    ----------

    data : `astropy.nddata.CCDData`
        CCD image on which the annuli are defined.

    annulus : `photutils.CircularAnnulus`
        One or more annulus (of any shape) from photutils.

    sigma : float, optional
        Number of standard deviations from the central value a
        pixel must be to reject it.

    iters : int, optional
        Maximum number of sigma clip iterations to perform. Iterations stop
        automatically if no pixels are rejected.

    Returns
    -------

    avg_sky_per_pix, med_sky_per_pix, std_sky_per_pix : `astropy.units.Quantity`
        Average, median and standard deviation of the sky per pixel.

    """
    # Get a list of masks from the annuli
    # Use the 'center' method because then pixels are either in or out. To use
    # 'partial' or 'exact' we would need to do a weighted sigma clip and
    # I'm not sure how to do that.
    masks = annulus.to_mask(method='center')

    anul_list = []
    for mask in masks:
        # Multiply the mask times the data
        to_clip = mask.multiply(data.data, fill_value=np.nan)
        anul_list.append(to_clip.flatten())
    # Convert the list to an array for doing the sigma clip
    anul_array = np.array(anul_list)
    # Turn all zeros into np.nan...
    anul_array[anul_array == 0] = np.nan
    avg_sky_per_pix, med_sky_per_pix, std_sky_per_pix = \
        faster_sigma_clip_stats(anul_array,
                                sigma=sigma,
                                iters=iters,
                                axis=1
                               )

    return (avg_sky_per_pix * data.unit, med_sky_per_pix * data.unit,
            std_sky_per_pix * data.unit)


def add_to_photometry_table(phot, ccd, annulus, apertures, fname='',
                            star_ids=None, camera=None,
                            bjd_coords=None, observatory_location=None,
                            fwhm_by_fit=True):
    """
    Calculate several columns for photometry table.

    Parameters
    ----------

    phot : `astropy.table.Table`
        An astropy Table with raw photometry data in it (generated by
        `photutils.aperture_photometry`).

    ccd : `astropy.nddata.CCDData`
        Image on which photometry is being done.

    annulus : `photutils.CircularAnnulus`
        One or more annulus (of any shape) from photutils.

    apertures : `photutils.CircularAperture`
        One or more apertures (of any shape) from photutils.

    fname : str, optional
        Name of the image file on which photometry is being performed.

    star_ids : str or int, optional
        ID for each of the sources.

    gain : float, optional
        Gain, in electrons/ADU, of the camera that took the image. The gain
        is used in calculating the instrumental magnitude.

    bjd_coords : `astropy.coordinates.SkyCoord`
        Coordinates of the object of interest in the Barycentric Julian Date
        frame. If not provided, the BJD column will not be added to the
        photometry table.

    observatory_location : str
        Name of the observatory where the images were taken. If not provided,
        the BJD column will not be added to the photometry table.

    fwhm_by_fit : bool, optional
        If ``True``, the FWHM will be calculated by fitting a Gaussian to
        the star. If ``False``, the FWHM will be calculated by finding the
        second moments of the light distribution. Default is ``True``.

    Returns
    -------

    None
        The input table is modified in place.
    """
    phot.rename_column('aperture_sum_0', 'aperture_sum')
    phot['aperture_sum'].unit = u.adu
    phot.rename_column('aperture_sum_1', 'annulus_sum')
    star_locs = ccd.wcs.all_pix2world(phot['xcenter'], phot['ycenter'], 0)
    star_coords = SkyCoord(ra=star_locs[0], dec=star_locs[1],
                           frame='icrs', unit='degree')
    phot['RA'] = star_coords.ra
    phot['Dec'] = star_coords.dec
    print('        ...calculating clipped sky stats')
    try:
        avg_sky_per_pix, med_sky_per_pix, std_sky_per_pix = \
            clipped_sky_per_pix_stats(ccd, annulus)
    except AttributeError:
        print('BAD ANNULUS!')
        avg_sky_per_pix, med_sky_per_pix, std_sky_per_pix = \
            np.nan, np.nan, np.nan
    print('        ...DONE calculating clipped sky stats')

    phot['sky_per_pix_avg'] = avg_sky_per_pix
    phot['sky_per_pix_med'] = med_sky_per_pix
    phot['sky_per_pix_std'] = std_sky_per_pix
    # Add width x/y:
    # Make small table with renamed columns
    fwhm_x, fwhm_y = compute_fwhm(ccd, phot, fit=fwhm_by_fit)

    # Set bad values to NaN now
    bad_fwhm = (fwhm_x < 0) | (fwhm_y < 0)
    fwhm_x[bad_fwhm] = np.nan
    fwhm_y[bad_fwhm] = np.nan

    phot['fwhm_x'] = fwhm_x
    phot['fwhm_y'] = fwhm_y
    phot['width'] = (fwhm_x + fwhm_y) / 2

    phot['aperture'] = apertures.r * u.pixel
    phot['aperture_area'] = apertures.area  # * u.pixel * u.pixel
    phot['annulus_inner'] = annulus.r_in * u.pixel
    phot['annulus_outer'] = annulus.r_out * u.pixel
    phot['annulus_area'] = annulus.area  # * u.pixel * u.pixel
    phot['exposure'] = [ccd.header['exposure']] * len(phot) * u.second
    phot['date-obs'] = [ccd.header['DATE-OBS']] * len(phot)
    night = Time(ccd.header['DATE-OBS'], scale='utc')
    night.format = 'mjd'
    phot['night'] = int(np.floor(night.value - 0.5))
    phot['aperture_net_flux'] = (phot['aperture_sum'] -
                                 (phot['aperture_area'] *
                                  phot['sky_per_pix_avg']))

    # This can happen, for example, when the object is faint
    # and centroiding is bad.
    bad_flux = phot['aperture_net_flux'] < 0
    all_bads = bad_flux | bad_fwhm

    phot['aperture_net_flux'][all_bads] = np.nan

    if observatory_location.lower() == 'feder':
        phot['BJD'] = find_times(phot['date-obs'][0], phot['exposure'][0],
                                 ra=bjd_coords.ra, dec=bjd_coords.dec)

    if camera is not None:
        phot['mag_inst'] = \
            (-2.5 * np.log10(camera.gain * phot['aperture_net_flux'].value /
                             phot['exposure'].value))

    metadata_to_add = ['AIRMASS', 'FILTER']
    for meta in metadata_to_add:
        phot[meta.lower()] = [ccd.header[meta]] * len(phot)
    if fname:
        phot['file'] = fname
    if star_ids is not None:
        phot['star_id'] = star_ids


def photometry_on_directory(directory_with_images, object_of_interest,
                            star_locs, aperture_rad,
                            inner_annulus, outer_annulus,
                            max_adu, star_ids,
                            camera,
                            bjd_coords=None,
                            observatory_location=None,
                            fwhm_by_fit=True):
    """
    Perform aperture photometry on a directory of images.

    Parameters
    ----------

    directory_with_images : str
        Folder containing the images on which to do photometry. Photometry
        will only be done on images that contain the ``object_of_interest``.

    object_of_interest : str
        Name of the object of interest. The only files on which photometry
        will be done are those whose header contains the keyword ``OBJECT``
        whose value is ``object_of_interest``.

    star_locs : tuple of numpy array
        The first entry in the tuple should be the right ascension of the
        sources, in degrees. The second should be the declination of
        the sources, in degrees.

    aperture_rad : int
        Radius of the aperture to use when performing photometry.

    inner_annulus : int
        Inner radius of annulus to use in for performing local sky
        subtraction.

    outer_annulus : int
        Outer radius of annulus to use in for performing local sky
        subtraction.

    max_adu : int
        Maximum allowed pixel value before a source is considered
        saturated.

    star_ids : array-like
        Unique identifier for each source in ``star_locs``.

    camera : `~stellarphot.Camera`
        Camera object which has gain, read noise and dark current set.

    gain : float
        Gain, in electrons/ADU, of the camera that took the image. The gain
        is used in calculating the instrumental magnitude.

    read_noise : float
        Read noise of the camera in electrons. Used in the CCD equation
        to calculate error.

    dark_current : float
        Dark current, in electron/sec. Used in the CCD equation
        to calculate error.

    bjd_coords : `astropy.coordinates.SkyCoord`
        Coordinates of the object of interest in the Barycentric Julian Date
        frame. If not provided, the BJD column will not be added to the
        photometry table.

    observatory_location : str
        Name of the observatory where the images were taken. If not provided,
        the BJD column will not be added to the photometry table.

    fwhm_by_fit : bool, optional
        If ``True``, the FWHM will be calculated by fitting a Gaussian to
        the star. If ``False``, the FWHM will be calculated by finding the
        second order moments of the light distribution. Default is ``True``.

    Returns
    -------

    phot : `astropy.table.Table`
        Table containing the photometry results.

    """
    ifc = ImageFileCollection(directory_with_images)
    phots = []
    missing_stars = []
    for a_ccd, fname in ifc.ccds(object=object_of_interest, return_fname=True):
        print('on image ', fname)
        try:
            # Convert RA/Dec to pixel coordinates for this image
            pix_coords = a_ccd.wcs.all_world2pix(star_locs[0], star_locs[1], 0)
        except AttributeError:
            print('    ....SKIPPING THIS IMAGE, NO WCS')
            continue

        xs, ys = pix_coords

        # Remove anything that is too close to the edges/out of frame
        padding = 3 * aperture_rad
        out_of_bounds = ((xs < padding) | (xs > (a_ccd.shape[1] - padding)) |
                         (ys < padding) | (ys > (a_ccd.shape[0] - padding)))
        in_bounds = ~out_of_bounds

        # Find centroids of each region around star that is in_bounds
        xs_in = xs[in_bounds]
        ys_in = ys[in_bounds]
        print('    ...finding centroids')
        try:
            xcen, ycen = centroid_sources(a_ccd.data, xs_in, ys_in,
                                          box_size=2 * aperture_rad + 1)
        except NoOverlapError:
            print('    ....SKIPPING THIS IMAGE, CENTROID FAILED')
            continue

        # Calculate offset between centroid in this image and the positions
        # based on input RA/Dec. Later we will set the magnitude of those with
        # large differences to an invalid value (maybe).
        center_diff = np.sqrt((xs_in - xcen)**2 + (ys_in - ycen)**2)

        # FWHM is typically 5-6 pixels. The center really shouldn't move
        # by more than that.
        too_much_shift = center_diff > 6

        xcen[too_much_shift] = xs_in[too_much_shift]
        ycen[too_much_shift] = ys_in[too_much_shift]

        # Set up apertures and annuli based on the centroids in this image.
        ap_locs = np.array([xcen, ycen]).T
        aps = CircularAperture(ap_locs, r=aperture_rad)

        anuls = CircularAnnulus(ap_locs, inner_annulus, outer_annulus)

        # Set any clearly bad values to NaN
        a_ccd.data[a_ccd.data > max_adu] = np.nan
        print('    ...doing photometry')
        # Do the photometry...
        pho = aperture_photometry(a_ccd.data, (aps, anuls),
                                  mask=a_ccd.mask, method='center')

        # We may have some stars we did not do photometry for because
        # those stars were out of bounds.
        # Add the ones we missed to the list of missing
        missed = star_ids[out_of_bounds]
        missing_stars.append(missed)

        # Add all the extra goodies to the table
        print('    ...adding extra columns')
        add_to_photometry_table(pho, a_ccd, anuls, aps,
                                fname=fname, star_ids=star_ids[in_bounds],
                                camera=camera, bjd_coords=bjd_coords,
                                observatory_location=observatory_location,
                                fwhm_by_fit=fwhm_by_fit)

        # And add the final table to the list of tables
        phots.append(pho)

    # ### Combine all of the individual photometry tables into one

    all_phot = vstack(phots)

    # ### Eliminate any stars that are missing from one or more images
    #
    # This makes life a little easier later...

    uniques = set()
    for miss in missing_stars:
        uniques.update(set(miss))

    actually_bad = sorted([u for u in uniques if u in all_phot['star_id']])
    len(uniques), len(actually_bad)

    all_phot.add_index('star_id')
    if actually_bad:
        bad_rows = all_phot.loc_indices[actually_bad]
        try:
            bad_rows = list(bad_rows)
        except TypeError:
            bad_rows = [bad_rows]
        all_phot.remove_indices('star_id')
        all_phot.remove_rows(sorted(bad_rows))

    all_phot.remove_indices('star_id')

    gain = camera.gain

    noise = calculate_noise(gain=camera.gain, read_noise=camera.read_noise,
                            dark_current_per_sec=camera.dark_current,
                            flux=all_phot['aperture_net_flux'],
                            sky_per_pix=all_phot['sky_per_pix_avg'].value,
                            aperture_area=all_phot['aperture_area'],
                            annulus_area=all_phot['annulus_area'],
                            exposure=all_phot['exposure'].value,
                            include_digitization=False)

    snr = gain * all_phot['aperture_net_flux'] / noise

    all_phot['mag_error'] = 1.085736205 / snr
    all_phot['noise'] = noise
    # AstroImageJ includes a factor of gain in the noise. IMHO it is part of the
    # flux but, for convenience, here it is
    all_phot['noise-aij'] = noise / gain
    all_phot['snr'] = snr

    return all_phot


def calculate_noise(gain=1.0, read_noise=0.0, dark_current_per_sec=0.0,
                    flux=0.0, sky_per_pix=0.0,
                    aperture_area=0, annulus_area=0,
                    exposure=0,
                    include_digitization=False):
    """
    Computes the noise in a photometric measurement.

    This function computes the noise (in units of gain) in a photometric measurement using the
    revised CCD equation from Collins et al (2017) AJ, 153, 77.  The equation is:

    .. math::

        \\sigma = \\sqrt{G \\cdot F + A \\cdot \\left(1 + \\frac{A}{B}\\right)\\cdot \\left[ G\\cdot S + D \\cdot t + R^2 + (0.289 G)^2\\right]}

    where :math:`\sigma` is the noise, :math:`G` is the gain, :math:`F` is the flux,
    :math:`A` is the aperture area in pixels, :math:`B` is the annulus area in pixels,
    :math:`S` is the sky per pixel, :math:`D` is the dark current per second,
    :math:`R` is the read noise, and :math:`t` is exposure time.

    Note: The :math:`(0.289 G)^2` term is "digitization noise" and is optional.

    Parameters
    ----------

    gain : float, optional
        Gain of the CCD. In units of electrons per DN.

    read_noise : float, optional
        Read noise of the CCD in electrons.

    dark_current_per_sec : float, optional
        Dark current of the CCD in electrons per second.

    flux : float, optional
        Flux of the source in electrons.

    sky_per_pix : float, optional
        Sky per pixel in electrons.

    aperture_area : int, optional
        Area of the aperture in pixels.

    annulus_area : int, optional
        Area of the annulus in pixels.

    exposure : int, optional
        Exposure time in seconds.

    include_digitization : bool, optional
        Whether to include the digitization noise. Defaults to False.

    Returns
    -------

    noise : float
        The noise in the photometric measurement.
    """

    try:
        no_annulus = (annulus_area == 0).all()
    except AttributeError:
        no_annulus = annulus_area == 0

    if no_annulus:
        area_ratio = aperture_area
    else:
        area_ratio = aperture_area * (1 + aperture_area / annulus_area)

    # Units in here are an absolute mess
    poisson_source = gain * flux

    try:
        poisson_source = poisson_source.value
    except AttributeError:
        pass

    sky = area_ratio * gain * sky_per_pix
    dark = area_ratio * dark_current_per_sec * exposure
    rn_error = area_ratio * read_noise ** 2

    digitization = 0.0

    if include_digitization:
        digitization = area_ratio * (gain * 0.289) ** 2

    return np.sqrt(poisson_source + sky + dark + rn_error + digitization)


def find_times(phot_column, exposure, ra, dec,
               latitude=46.86678, longitude=263.54672):
    """
    Returns a numpy array of barycentric Julian date times

    Parameters
    ----------

    phot_column : `astropy.table.Column` or numpy array
        numpy array or column of observation dates from the photometry table

    exposure : float, optional
        exposure time in seconds

    RA : float
        Right ascension in degree units

    Dec : float
        Declination  in degree units

    latitude : float, optional
        latitude of the observatory in degrees North, default is for Paul P. Feder Observatory

    longitude : float, optional
        longitude of the observatory in degree East of Greenwich (0 to 360), default
        is for Paul P. Feder Observatory


    Returns
    -------

    numpy array
        array of times in barycentric Julian date

    """
    location = EarthLocation(lat=latitude, lon=longitude)

    times = Time(phot_column, scale='utc', format='isot', location=location)
    ip_peg = SkyCoord(ra=ra, dec=dec, unit='degree')
    ltt_bary = times.light_travel_time(ip_peg)
    times_tdb = times.tdb
    time_barycenter = times_tdb + ltt_bary

    # Adjust to midpoint of exposure
    bary_time = time_barycenter + exposure / 2

    return bary_time.jd


from pathlib import Path

import numpy as np

from astropy.coordinates import SkyCoord
from astropy.table import Table
from astropy.utils.data import get_pkg_data_filename

from stellarphot.io import ApertureFileAIJ
from stellarphot.io.aij import ApertureAIJ


def test_aperture_eq():
    ap1 = ApertureAIJ()
    ap2 = ApertureAIJ()
    ap3 = ApertureAIJ()
    ap3.rback1 = 27.0
    assert ap1 == ap2
    assert ap2 == ap3


def test_aperture_file_content():
    # Test that the format of the generated aperture file
    # matches the expected format
    ap = ApertureFileAIJ()
    ref_data = get_pkg_data_filename('data/apertures_as_table.csv')
    ref_table = Table.read(ref_data)

    ap.multiaperture.xapertures = ref_table['x']

    # AIJ has origin in different place than the reference table.
    ap.multiaperture.yapertures = np.around((4096 - ref_table['y']),
                                            decimals=4)

    ap.multiaperture.raapertures = ref_table['ra']
    ap.multiaperture.decapertures = ref_table['dec']

    ap.multiaperture.isrefstar = ref_table['isrefstar']
    ap.multiaperture.centroidstar = ref_table['centroidstar']
    ap.multiaperture.isalignstar = ref_table['isalignstar']

    ap.multiaperture.absmagapertures = ref_table['absmag']

    ref_aperture_file = \
        get_pkg_data_filename('data/aij-sample-apertures.aperture')

    ref_apertures = ApertureFileAIJ.read(ref_aperture_file)

    assert ref_apertures == ap


def test_aperture_creation_from_table():
    # Check that generating an aperture object from an
    # aperture table gives the right result.

    ref_data = get_pkg_data_filename('data/apertures_as_table.csv')
    ref_table = Table.read(ref_data)

    # Need to create a coord column
    coordinates = SkyCoord(ra=ref_table['ra'], dec=ref_table['dec'],
                           unit='degree')
    ref_table['coord'] = coordinates

    # Delete some columns that are not in the usual aperture table
    del ref_table['ra'], ref_table['dec'], \
        ref_table['isalignstar'], ref_table['centroidstar']

    # Generate marker names that match those in aperture file
    # Note that the csv reads in the True/False column as text,
    # not as bool.
    ref_table['marker name'] = ['APASS comparison'
                                if v == 'True' else "TESS target"
                                for v in ref_table['isrefstar']]

    del ref_table['isrefstar']

    ap_info = ApertureAIJ()
    ap_aij = ApertureFileAIJ.from_table(ref_table, aperture_rad=ap_info.radius,
                                        inner_annulus=ap_info.rback1,
                                        outer_annulus=ap_info.rback2,
                                        )
    ref_aperture_file = \
        get_pkg_data_filename('data/aij-sample-apertures.aperture')

    ref_apertures = ApertureFileAIJ.read(ref_aperture_file)

    assert ref_apertures == ap_aij



from importlib import resources
import yaml

import pytest

from astropy.utils.data import get_pkg_data_filename
from astropy.table import Table
from stellarphot.io import AAVSOExtendedFileFormat, AAVSOExtendedFileFormatColumns

DEFAULT_OBSCODE = "ABCDE"


def test_no_obscode_raises_error():
    with pytest.raises(TypeError, match="obscode"):
        aef = AAVSOExtendedFileFormat()


def test_default_values():
    aef = AAVSOExtendedFileFormat(DEFAULT_OBSCODE)
    assert aef.delim == ","
    assert len(aef.magnitude) == 0


def test_setting_type_raises_error():
    aef = AAVSOExtendedFileFormat(DEFAULT_OBSCODE)
    with pytest.raises(AttributeError, match="(no setter)|(can't set attribute)"):
        aef.type = 'STD'


def set_up_aef():
    aef = AAVSOExtendedFileFormat(DEFAULT_OBSCODE)
    aef_cols = AAVSOExtendedFileFormatColumns
    test_file = get_pkg_data_filename('data/dy-her-aavso-test-data.csv')
    input_data = Table.read(test_file)
    var_star_name = "dy her"
    var_star_id = 1
    # Data here should be the calibrated magnitude and error
    aef.set_data_columns(input_data,
                         {'mag': aef_cols.VARIABLE_MAG,
                          'mag_error': aef_cols.VARIABLE_MAG_ERROR,
                          'airmass': aef_cols.AIRMASS,
                          "jd_mid_utc": aef_cols.JD,},
                         star_id=var_star_id)
    aef.starid = var_star_name
    aef.filter = "rp"

    # If the method is NOT ensemble, then report check star instrumental magnitude.
    check_star_auid = "000-BBX-643"
    check_star_id = 63
    aef.set_data_columns(input_data,
                         {'mag_inst': aef_cols.CHECK_STAR_MAG},
                         star_id=check_star_id)
    aef.kname = check_star_auid

    # Instrumental mag of the comparison star unless ensemble.
    comp_star_auid = "000-BJR-400"
    comp_star_id = 26
    aef.set_data_columns(input_data,
                         {'mag_inst': aef_cols.COMP_STAR_MAG},
                         star_id=comp_star_id)
    aef.cname = comp_star_auid
    return aef, input_data


def test_making_table():
    aef, input_data = set_up_aef()
    aef_table = aef.to_table()
    check_star_auid = "000-BBX-643"
    comp_star_auid = "000-BJR-400"
    var_star_name = "dy her"
    assert set(aef_table["KNAME"]) == set([check_star_auid])
    assert set(aef_table["CNAME"]) == set([comp_star_auid])
    assert set(aef_table["STARID"]) == set([var_star_name])
    assert set(aef_table["FILTER"]) == set(["rp"])
    assert len(aef_table) == len(set(input_data['BJD']))


def test_making_table_ensemble():
    aef, input_data = set_up_aef()
    aef.ensemble = True
    assert set(aef.to_table()["CMAG"]) == set(["na"])

    with pytest.raises(ValueError, match="Cannot set comparison star"):
        aef.set_data_columns(input_data,
                             {'mag_inst': AAVSOExtendedFileFormatColumns.COMP_STAR_MAG},
                              star_id=26)


@pytest.mark.parametrize("ensemble", [True, False])
def test_table_column_formats(tmp_path, ensemble):
    aef, input_data = set_up_aef()
    aef.ensemble = ensemble
    aef_table = aef.to_table()
    aef_table.write(tmp_path / "test.csv")
    with open(tmp_path / "test.csv") as f:
        lines = f.readlines()

    # Get the file specification
    table_description = resources.read_text('stellarphot.io', 'aavso_submission_schema.yml')
    table_structure = yaml.safe_load(table_description)
    values = []
    columns = []
    errors = {}
    for line in lines:
        if line.startswith('#'):
            print("#")
            continue

        if str.isalpha(line[0]) and not columns:
            columns = line.strip().split(",")
            print(f'{columns=}')
            continue

        values = line.strip().split(",")

        for column, value in zip(columns, values):
            table_structure['data'].keys()
            length_limit = table_structure['data'][column].get('limit', None)
            if length_limit:
                try:
                    assert len(value) <= length_limit
                except AssertionError as e:
                    errors[column] = str(e)
        break

    error_messages = '\n'.join([f'Column {error[0]}: {error[1]}' for error in errors.items()])

    if error_messages:
        raise AssertionError(f"Errors in table format:\n{error_messages}")



import functools
from pathlib import Path

import pandas

import ipywidgets as ipw

import numpy as np

from astropy.table import Table
from astropy.coordinates import SkyCoord
from astropy import units as u
from astropy.nddata import CCDData
from astropy import units as u
from astropy.coordinates.name_resolve import NameResolveError


try:
    from astrowidgets import ImageWidget
except ImportError:
    from astrowidgets.ginga import ImageWidget

from stellarphot.differential_photometry import *
from stellarphot.io import TessSubmission, TOI, TessTargetFile
from stellarphot.photometry import *
from stellarphot.visualization.seeing_profile_functions import set_keybindings
from stellarphot.visualization.fits_opener import FitsOpener


__all__ = ['read_file', 'set_up', 'crossmatch_APASS2VSX', 'mag_scale',
           'in_field', 'make_markers', 'wrap', 'ComparisonViewer']

DESC_STYLE = {"description_width": "initial"}


def read_file(radec_file):
    """
    Read an AIJ radec file with target and/or comp positions

    Parameters
    ----------

    radec_file : str
        Name of the file

    Returns
    -------

    `astropy.table.Table`
        Table with target information, including a
        `astropy.coordinates.SkyCoord` column.
    """
    df = pandas.read_csv(radec_file, names=['RA', 'Dec', 'a', 'b', 'Mag'])
    target_table = Table.from_pandas(df)
    ra = target_table['RA']
    dec = target_table['Dec']
    target_table['coords'] = SkyCoord(ra=ra, dec=dec, unit=(u.hour, u.degree))
    return target_table


def set_up(sample_image_for_finding_stars, directory_with_images='.'):
    """
    Read in sample image and find known variables in the field of view.

    Parameters
    ----------

    sample_image_for_finding_stars : str
        Name or URL of a FITS image of the field of view.

    directory_with_images : str, optional
        Folder in which the image is located. Ignored if the sample image
        is a URL.

    Returns
    -------

    ccd: `astropy.nddata.CCDData`
        Sample image.

    vsx: `astropy.table.Table`
        Table with known variables in the field of view.

    """
    if sample_image_for_finding_stars.startswith('http'):
        path = sample_image_for_finding_stars
    else:
        path = Path(directory_with_images) / sample_image_for_finding_stars

    ccd = CCDData.read(path)
    try:
        vsx = find_known_variables(ccd)
    except RuntimeError:
        vsx = []
    else:
        ra = vsx['RAJ2000']
        dec = vsx['DEJ2000']
        vsx['coords'] = SkyCoord(ra=ra, dec=dec, unit=(u.hour, u.degree))

    return ccd, vsx


def crossmatch_APASS2VSX(CCD, RD, vsx):
    """
    Find APASS stars in FOV and matches APASS stars to VSX and APASS to input targets.

    Parameters
    ----------

    CCD : `astropy.nddata.CCDData`
        Sample image.

    RD : `astropy.table.Table`
        Table with target information, including a
        `astropy.coordinates.SkyCoord` column.

    vsx : `astropy.table.Table`
        Table with known variables in the field of view.

    Returns
    -------

    apass : `astropy.table.Table`
        Table with APASS stars in the field of view.

    v_angle : `astropy.units.Quantity`
        Angular separation between APASS stars and VSX stars.

    RD_angle : `astropy.units.Quantity`
        Angular separation between APASS stars and input targets.
    """
    apass, apass_in_bright = find_apass_stars(CCD)
    ra = apass['RAJ2000']
    dec = apass['DEJ2000']
    apass['coords'] = SkyCoord(ra=ra, dec=dec, unit=(u.hour, u.degree))
    apass_coord = apass['coords']

    if vsx:
        v_index, v_angle, v_dist = \
            apass_coord.match_to_catalog_sky(vsx['coords'])
    else:
        v_angle = []

    if RD:
        RD_index, RD_angle, RD_dist = \
            apass_coord.match_to_catalog_sky(RD['coords'])
    else:
        RD_angle = []

    return apass, v_angle, RD_angle


def mag_scale(cmag, apass, v_angle, RD_angle,
              brighter_dmag=0.44, dimmer_dmag=0.75):
    """
    Select comparison stars that are 1) not close the VSX stars or to other
    target stars and 2) fall within a particular magnitude range.

    Parameters
    ----------

    cmag : float
        Magnitude of the target star.

    apass : `astropy.table.Table`
        Table with APASS stars in the field of view.

    v_angle : `astropy.units.Quantity`
        Angular separation between APASS stars and VSX stars.

    RD_angle : `astropy.units.Quantity`
        Angular separation between APASS stars and input targets.

    brighter_dmag : float, optional
        Maximum difference in magnitude between the target and comparison stars.

    dimmer_dmag : float, optional
        Minimum difference in magnitude between the target and comparison stars.

    Returns
    -------

    apass_good_coord : `astropy.coordinates.SkyCoord`
        Coordinates of the comparison stars.

    good_stars : `astropy.table.Table`
        Table with the comparison stars.
    """
    high_mag = apass['r_mag'] < cmag + dimmer_dmag
    low_mag = apass['r_mag'] > cmag - brighter_dmag
    if len(v_angle) > 0:
        good_v_angle = v_angle > 1.0 * u.arcsec
    else:
        good_v_angle = True

    if len(RD_angle) > 0:
        good_RD_angle = RD_angle > 1.0 * u.arcsec
    else:
        good_RD_angle = True

    good_stars = high_mag & low_mag & good_RD_angle & good_v_angle
    good_apass = apass[good_stars]
    apass_good_coord = good_apass['coords']
    return apass_good_coord, good_stars


def in_field(apass_good_coord, ccd, apass, good_stars):
    """
    Return APASS stars in the field of view.

    Parameters
    ----------

    apass_good_coord : `astropy.coordinates.SkyCoord`
        Coordinates of the comparison stars.

    ccd : `astropy.nddata.CCDData`
        Sample image.

    apass : `astropy.table.Table`
        Table with APASS stars in the field of view.

    good_stars : `astropy.table.Table`
        Table with the comparison stars.

    Returns
    -------

    ent : `astropy.table.Table`
        Table with APASS stars in the field of view.
    """
    apassx, apassy = ccd.wcs.all_world2pix(
        apass_good_coord.ra, apass_good_coord.dec, 0)
    ccdx, ccdy = ccd.shape

    xin = (apassx < ccdx) & (0 < apassx)
    yin = (apassy < ccdy) & (0 < apassy)
    xy_in = xin & yin
    apass_good_coord[xy_in]
    nt = apass[good_stars]
    ent = nt[xy_in]
    return ent


def make_markers(iw, ccd, RD, vsx, ent, name_or_coord=None):
    """
    Add markers for APASS, TESS targets, VSX.  Also center on object/coordinate.

    Parameters
    ----------

    iw : `astrowidgets.ImageWidget`
        Ginga widget.

    ccd : `astropy.nddata.CCDData`
        Sample image.

    RD : `astropy.table.Table`
        Table with target information, including a
        `astropy.coordinates.SkyCoord` column.

    vsx : `astropy.table.Table`
        Table with known variables in the field of view.

    ent : `astropy.table.Table`
        Table with APASS stars in the field of view.

    name_or_coord : str or `astropy.coordinates.SkyCoord`, optional
        Name or coordinates of the target.

    Returns
    -------

    None
        Markers are added to the image in Ginga widget.
    """
    iw.load_nddata(ccd)
    iw.zoom_level = 'fit'

    try:
        iw.reset_markers()
    except AttributeError:
        iw.remove_all_markers()

    if RD:
        iw.marker = {'type': 'circle', 'color': 'green', 'radius': 10}
        iw.add_markers(RD, skycoord_colname='coords',
                       use_skycoord=True, marker_name='TESS Targets')

    if name_or_coord is not None:
        if isinstance(name_or_coord, str):
            iw.center_on(SkyCoord.from_name(name_or_coord))
        else:
            iw.center_on(name_or_coord)

    if vsx:
        iw.marker = {'type': 'circle', 'color': 'blue', 'radius': 10}
        iw.add_markers(vsx, skycoord_colname='coords',
                       use_skycoord=True, marker_name='VSX')
    iw.marker = {'type': 'circle', 'color': 'red', 'radius': 10}
    iw.add_markers(ent, skycoord_colname='coords',
                   use_skycoord=True, marker_name='APASS comparison')
    iw.marker = {'type': 'cross', 'color': 'red', 'radius': 6}


def wrap(imagewidget, outputwidget):
    """
    Utility function to let you click to select/deselect comparisons.

    Parameters
    ----------

    imagewidget : `astrowidgets.ImageWidget`
        Ginga widget.

    outputwidget : `ipywidgets.Output`
        Output widget for printing information.

    """
    def cb(viewer, event, data_x, data_y):
        i = imagewidget._viewer.get_image()

        try:
            imagewidget.next_elim += 1
        except AttributeError:
            imagewidget.next_elim = 1
        pad = 15
        x = int(np.floor(event.data_x))
        y = int(np.floor(event.data_y))
        ra, dec = i.wcs.wcs.all_pix2world(event.data_x, event.data_y, 0)
        out_skycoord = SkyCoord(ra=ra, dec=dec, unit=(u.degree, u.degree))

        try:
            all_table = imagewidget.get_markers(marker_name='all')
        except AttributeError:
            all_table = imagewidget.get_all_markers()

        with outputwidget:
            index, d2d, d3d = out_skycoord.match_to_catalog_sky(
                all_table['coord'])
            if d2d < 10 * u.arcsec:
                mouse = all_table['coord'][index].separation(
                    all_table['coord'])
                rat = mouse < 1 * u.arcsec
                elims = [name for name in all_table['marker name']
                         [rat] if name.startswith('elim')]
                if not elims:
                    imagewidget.add_markers(all_table[rat], skycoord_colname='coord', use_skycoord=True,
                                            marker_name=f'elim{imagewidget.next_elim}')
                else:
                    for elim in elims:
                        try:
                            imagewidget.remove_markers_by_name(marker_name=elim)
                        except AttributeError:
                            imagewidget.remove_markers(marker_name=elim)

            else:
                print('sorry try again')
                imagewidget._viewer.onscreen_message('Click closer to a star')

    return cb


class ComparisonViewer:
    """
    A class to store an instance of the comparison viewer.

    Attributes
    ----------

    target_mag : float
        Magnitude of the target.

    bright_mag_limit : float
        Bright magnitude limit for APASS stars.

    dim_mag_limit : float
        Dim magnitude limit for APASS stars.

    targets_from_file : str
        File with target information.

    tess_submission : `~stellarphot.io.TessSubmission`
        Instance of the TESS submission class.

    target_coord : `astropy.coordinates.SkyCoord`
        Coordinates of the target.

    box : `ipywidgets.Box`
        Box containing the widgets.

    iw : `ginga.util.grc.RemoteClient`
        Ginga widget.

    aperture_output_file : str
        File to save aperture information to.
    """
    def __init__(self,
                 file="",
                 directory='.',
                 target_mag=10,
                 bright_mag_limit=8,
                 dim_mag_limit=17,
                 targets_from_file=None,
                 object_coordinate=None,
                 aperture_output_file=None):
        """
        Initializes an instance of the ComparisonViewer class.

        Parameters
        ----------

        file : str, optional
            File to open. Defaults to "".

        directory : str, optional
            Directory to open file from. Defaults to '.'.

        target_mag : float, optional
            Magnitude of the target. Defaults to 10.

        bright_mag_limit : float, optional
            Bright magnitude limit for APASS stars. Defaults to 8.

        dim_mag_limit : float, optional
            Dim magnitude limit for APASS stars.  Defaults to 17.

        targets_from_file : str, optional
            File with target information.  Defaults to None.

        object_coordinate : `astropy.coordinates.SkyCoord`, optional
            Coordinates of the target. Defaults to None.

        aperture_output_file : str, optional
            File to save aperture information to.  Defaults to None.
        """

        self._label_name = 'labels'
        self._circle_name = 'target circle'
        self._file_chooser = FitsOpener()

        self._directory = directory
        self.target_mag = target_mag
        self.bright_mag_limit = bright_mag_limit
        self.dim_mag_limit = dim_mag_limit
        self.targets_from_file = targets_from_file
        self.tess_submission = None
        self._tess_object_info = None
        self.target_coord = object_coordinate

        self.box, self.iw = self._viewer()

        self.aperture_output_file = aperture_output_file

        if file:
            self._file_chooser.set_file(file, directory=directory)
            self._set_file(None)

        self._make_observers()

    def _init(self):
        """
        Handles aspects of initialization that need to be defered until a file is chosen.
        """
        if self.tess_submission is not None:
            self._tess_object_info.layout.visibility = "visible"
        self.ccd, self.vsx = \
            set_up(self._file_chooser.path.name,
                   directory_with_images=self._file_chooser.path.parent
                   )

        apass, vsx_apass_angle, targets_apass_angle = crossmatch_APASS2VSX(self.ccd,
                                                                            self.targets_from_file,
                                                                            self.vsx)

        apass_good_coord, good_stars = mag_scale(self.target_mag, apass, vsx_apass_angle,
                                                 targets_apass_angle,
                                                 brighter_dmag=self.target_mag - self.bright_mag_limit,
                                                 dimmer_dmag=self.dim_mag_limit - self.target_mag)

        apass_comps = in_field(apass_good_coord, self.ccd, apass, good_stars)
        make_markers(self.iw, self.ccd, self.targets_from_file, self.vsx, apass_comps,
                     name_or_coord=self.target_coord)

    @property
    def variables(self):
        """
        Return a dictionary of the variables in the class.

        Returns
        -------

        our_vsx : `astropy.table.Table`
            Table of the variables in the class.

        """
        comp_table = self.generate_table()
        new_vsx_mark = comp_table['marker name'] == 'VSX'
        idx, _, _ = comp_table['coord'][new_vsx_mark].match_to_catalog_sky(self.vsx['coords'])
        our_vsx = self.vsx[idx]
        our_vsx['star_id'] = comp_table['star_id'][new_vsx_mark]
        return our_vsx

    def _set_object(self):
        """
        Try to automatically set object name immediately after file is chosen.
        """
        try:
            self.object_name.value = self._file_chooser.header['object']
        except KeyError:
            # No object, will show empty box for name
            self.object_name.disabled = False

        # We have a name, try to get coordinates from it
        try:
            self.target_coord = SkyCoord.from_name(self.object_name.value)
        except NameResolveError:
            pass

        # Maybe this is a tess object?
        try:
            self.tess_submission = TessSubmission.from_header(
                self._file_chooser.header,
                telescope_code="Paul-P-Feder-0.4m",
                planet=1
            )
        except ValueError:
            # Guess not, time to turn on the coordinates box
            # self._turn_on_coordinates()
            self.tess_submission = None
            self.toi_info = None
            self.targets_from_file = None
            self._tess_object_info.layout.visibility = "hidden"
            self.tess_save_toggle.value = False
            self.tess_save_toggle.disabled = True
        else:
            self.tess_save_toggle.disabled = False
            self.toi_info = TOI(self.tess_submission.tic_id)

            self._target_file_info = TessTargetFile(self.toi_info.coord,
                                                    self.toi_info.tess_mag,
                                                    self.toi_info.depth)

            self.target_coord = self.tess_submission.tic_coord
            self._tess_object_info.mag.value = self.toi_info.tess_mag
            self._tess_object_info.depth.value = self.toi_info.depth
            self._tess_object_info.layout.visibility = "visible"
            self.targets_from_file = self._target_file_info.table

    def _set_file(self, change):
        self._set_object()
        self._init()
        self._update_tess_save_names()

    def _save_toggle_action(self, change):
        activated = change['new']

        if activated:
            self._tess_save_box.layout.visibility = "visible"
        else:
            self._tess_save_box.layout.visibility = "hidden"

    def _make_observers(self):
        self._show_labels_button.observe(self._show_label_button_handler,
                                         names='value')
        self._save_var_info.on_click(self._save_variables_to_file)
        self._save_aperture_file.on_click(self._save_aperture_to_file)
        self._file_chooser.register_callback(self._set_file)
        self.tess_save_toggle.observe(self._save_toggle_action, "value")
        self.save_files.on_click(self.save_tess_files)

    def _save_variables_to_file(self, button=None, filename=''):
        if not filename:
            filename = 'variables.csv'
        self.variables.write(filename)

    def _show_label_button_handler(self, change):
        value = change['new']
        if value:
            # Showing labels can take a bit, disable while in progress
            self._show_labels_button.disabled = True
            self.show_labels()
            self._show_labels_button.disabled = False
        else:
            self.remove_labels()
        self._show_labels_button.description = self._show_labels_button.descriptions[value]

    def _save_aperture_to_file(self, button=None, filename=''):
        if not filename:
            filename = self.aperture_output_file

        self.generate_table().write(filename)

    def _make_control_bar(self):
        self._show_labels_button = ipw.ToggleButton(description='Click to show labels')
        self._show_labels_button.descriptions = {
            True: 'Click to hide labels',
            False: 'Click to show labels'
        }

        self._save_var_info = ipw.Button(description='Save variable info')

        self._save_aperture_file = ipw.Button(description='Save aperture file')

        controls = ipw.HBox(
            children=[
                self._show_labels_button,
                self._save_var_info,
                self._save_aperture_file
            ]
        )

        return controls

    def _make_tess_object_info(self):
        """
        Make the controls for the TESS mag and depth used to look up GAIA target list.
        """
        tess_object_info = ipw.HBox()
        tess_mag = ipw.FloatText(description="TESS mag")
        tess_depth = ipw.FloatText(description="Depth (ppt)")
        tess_object_info.children = [tess_mag, tess_depth]
        self._tess_object_info = tess_object_info
        self._tess_object_info.mag = tess_mag
        self._tess_object_info.depth = tess_depth
        if self.tess_submission is None:
            self._tess_object_info.layout.visibility = "hidden"

    def _make_tess_save_box(self):
        self.tess_save_toggle = ipw.ToggleButton(description="TESS files...",
                                            disabled=True)
        self._tess_save_box = ipw.VBox()
        self._tess_save_box.layout.visibility = "hidden"

        scope_name = ipw.Text(description="Telescope code",
                              value="Paul-P-Feder-0.4m",
                              style=DESC_STYLE)

        planet_num = ipw.IntText(description="Planet", value=1)

        dumb = []
        dumb2 = []

        for save in ["Full field of view", "Zoomed filed of view"]:  # , "Aperture file"]:
            box = ipw.HBox()
            title = ipw.HTML(value=f"<b>{save} file name</b>")
            label = ipw.Label(value="")
            box.children = (title, label)
            dumb.append(label)
            dumb2.append(box)

        # self._field_name, self._zoom_name, self._aper_name = dumb
        self._field_name, self._zoom_name = dumb
        self.save_files = ipw.Button(description="Save")
        self._tess_save_box.children = [scope_name, planet_num] + dumb2 + [self.save_files]

    def _update_tess_save_names(self):
        if self.tess_submission is not None:
            self._field_name.value = self.tess_submission.field_image
            self._zoom_name.value = self.tess_submission.field_image_zoom
            # self._aper_name.value = self.tess_submission.apertures
        else:
            self._field_name.value = ""
            self._zoom_name.value = ""

    def _viewer(self):
        header = ipw.HTML(value="""
        <h2>Click and drag or use arrow keys to pan, use +/- keys to zoom</h2>
        <h3>Shift-left click (or Crtl-left click)to exclude star as target
        or comp. Click again to include.</h3>
        """)

        legend = ipw.HTML(value="""
        <h3>Green circles -- Gaia stars within 2.5 arcmin of target</h3>
        <h3>Red circles -- APASS stars within 1 mag of target</h3>
        <h3>Blue circles -- VSX variables</h3>
        <h3>Red × -- Exclude as target or comp</h3>
        """)

        iw = ImageWidget()
        out = ipw.Output()
        set_keybindings(iw)
        bind_map = iw._viewer.get_bindmap()
        gvc = iw._viewer.get_canvas()
        bind_map.map_event(None, ('shift',), 'ms_left', 'cursor')
        gvc.add_callback('cursor-down', wrap(iw, out))

        self.object_name = ipw.Text(description="Object", disabled=True)
        controls = self._make_control_bar()
        self._make_tess_object_info()
        self._make_tess_save_box()
        box = ipw.VBox()
        inner_box = ipw.HBox()
        inner_box.children = [iw, legend]
        box.children = [self._file_chooser.file_chooser, self.object_name, self._tess_object_info, header,
                        inner_box, controls, self.tess_save_toggle, self._tess_save_box]

        return box, iw

    def save_tess_files(self, button=None):
        """
        Save the TESS files.

        Parameters
        ----------

        button : `ipywidgets.Button`, optional
            The button that was clicked.

        Returns
        -------

        None
            Button to save TESS file set to true (triggering action).
        """
        if self._field_name.value:
            self.tess_field_view()
            self.iw.save(self._field_name.value, overwrite=True)

        if self._zoom_name.value:
            self.tess_field_zoom_view()
            self.iw.save(self._zoom_name.value, overwrite=True)

    def generate_table(self):
        """
        Generate the table of stars to use for the aperture file.

        Returns
        -------
        comp_table : `astropy.table.Table`
            Table of stars to use for the aperture file.
        """
        try:
            all_table = self.iw.get_all_markers()
        except AttributeError:
            all_table = self.iw.get_markers(marker_name='all')

        elims = np.array([name.startswith('elim')
                         for name in all_table['marker name']])
        elim_table = all_table[elims]
        comp_table = all_table[~elims]

        index, d2d, d3d = elim_table['coord'].match_to_catalog_sky(comp_table['coord'])
        comp_table.remove_rows(index)

        # Add separate RA and Dec columns for ease in processing later
        comp_table['ra'] = comp_table['coord'].ra.degree
        comp_table['dec'] = comp_table['coord'].dec.degree

        # Calculate how far each is from target
        comp_table['separation'] = self.target_coord.separation(comp_table['coord'])

        # Add dummy column for sorting in the order we want
        comp_table['sort'] = np.zeros(len(comp_table))

        # Set sort order
        apass_mark = comp_table['marker name'] == 'APASS comparison'
        vsx_mark = comp_table['marker name'] == 'VSX'
        tess_mark = ((comp_table['marker name'] == 'TESS Targets') |
                (comp_table['separation'] < 0.3 * u.arcsec))


        comp_table['sort'][apass_mark] = 2
        comp_table['sort'][vsx_mark] = 1
        comp_table['sort'][tess_mark] = 0

        # Sort the table
        comp_table.sort(['sort', 'separation'])

        # Assign the IDs
        comp_table['star_id'] = range(1, len(comp_table) + 1)

        return comp_table

    def show_labels(self, marker_table=None):
        """
        Show the labels for the stars.

        Parameters
        ----------

        marker_table : `astropy.table.Table`, optional
            Table of stars to use for the aperture file -- this overrides whatever
            markers are already in the viewer.

        Returns
        -------

        None
            Labels for the stars are shown.
        """
        plot_names = []

        if marker_table is None:
            comp_table = self.generate_table()
        else:
            comp_table = marker_table

        original_mark = self.iw._marker
        for star in comp_table:
            star_id = star['star_id']
            if star['marker name'] == 'TESS Targets':
                label = f'T{star_id}'
                self.iw._marker = functools.partial(self.iw.dc.Text, text=label, fontsize=20, fontscale=False, color='green')
                self.iw.add_markers(Table(data=[[star['x']+20], [star['y']-20]], names=['x', 'y']),
                               marker_name=self._label_name)

            elif star['marker name'] == 'APASS comparison':
                label = f'C{star_id}'
                self.iw._marker = functools.partial(self.iw.dc.Text, text=label, fontsize=20, fontscale=False, color='red')
                self.iw.add_markers(Table(data=[[star['x']+20], [star['y']-20]], names=['x', 'y']),
                                    marker_name=self._label_name)

            elif star['marker name'] == 'VSX':
                label = f'V{star_id}'
                self.iw._marker = functools.partial(self.iw.dc.Text, text=label, fontsize=20, fontscale=False, color='blue')
                self.iw.add_markers(Table(data=[[star['x']+20], [star['y']-20]], names=['x', 'y']),
                                    marker_name=self._label_name)
            else:
                label = f'U{star_id}'
                print(f"Unrecognized marker name: {star['marker name']}")
            plot_names.append(label)
        self.iw._marker = original_mark

    def remove_labels(self):
        """
        Remove the labels for the stars.

        Returns
        -------

        None
            Labels for the stars are removed.
        """
        try:
            try:
                self.iw.remove_markers(marker_name=self._label_name)
            except AttributeError:
                self.iw.remove_markers_by_name(marker_name=self._label_name)
        except ValueError:
            # No labels, keep going
            pass

    def show_circle(self,
                    radius=2.5 * u.arcmin,
                    pixel_scale=0.56 * u.arcsec / u.pixel):
        """
        Show a circle around the target.

        Parameters
        ----------

        radius : `astropy.units.Quantity`, optional
            Radius of circle. The default is ``2.5*u.arcmin``.

        pixel_scale : `astropy.units.Quantity`, optional
            Pixel scale of image. The default is ``0.56*u.arcsec/u.pixel``.

        Returns
        -------

        None
            Circle is shown around the target.
        """
        radius_pixels = np.round((radius / pixel_scale).to(u.pixel).value,
                                 decimals=0)
        orig_marker = self.iw.marker
        self.iw.marker = {'color': 'yellow',
                          'radius': radius_pixels,
                          'type': 'circle'}
        self.iw.add_markers(Table(data=[[self.target_coord]], names=['coords']),
                            skycoord_colname='coords',
                            use_skycoord=True, marker_name=self._circle_name)
        self.iw.marker = orig_marker

    def remove_circle(self):
        """
        Remove the circle around the target.

        Returns
        -------

        None
            Circle is removed from the image.
        """
        try:
            self.iw.remove_markers(marker_name=self._circle_name)
        except AttributeError:
            self.iw.remove_markers_by_name(marker_name=self._circle_name)

    def tess_field_view(self):
        """
        Show the whole TESS field of view including circle around target, but hide labels.

        Returns
        -------

        None
            Shows image as described above.
        """

        # Show whole field of view
        self.iw.zoom_level = 'fit'

        # Show the circle
        self.show_circle()

        # Turn off labels -- too cluttered
        self.remove_labels()

    def tess_field_zoom_view(self, width=6 * u.arcmin):
        """
        Zoom in on the TESS field of view.

        Parameters
        ----------

        width : `astropy.units.Quantity`, optional
            Width of field of view. The default is ``6*u.arcmin``.

        Returns
        -------

        None
            Zooms in on the image as described above.
        """

        # Turn off labels -- too cluttered
        self.remove_labels()

        left_side = self.ccd.wcs.pixel_to_world(0, self.ccd.shape[1]/2)
        right_side = self.ccd.wcs.pixel_to_world(self.ccd.shape[0], self.ccd.shape[1]/2)
        fov = left_side.separation(right_side)

        view_ratio = width / fov
        # fit first to get zoom level at full field of view
        self.iw.zoom_level = "fit"

        # Then set it to what we actually want...
        self.iw.zoom_level = self.iw.zoom_level / view_ratio

        # Show the circle
        self.show_circle()



import numpy as np
import pytest

from astropy.table import Table
from astropy.coordinates import SkyCoord
from astropy.utils.data import get_pkg_data_filename
from astropy.io import fits
from astropy.wcs import WCS
from astropy.nddata import CCDData

from ..catalog_search import catalog_clean, in_frame, \
                             catalog_search, find_known_variables, \
                             find_apass_stars, filter_catalog
from ...tests.make_wcs import make_wcs

CCD_SHAPE = [2048, 3073]


def a_table(masked=False):
    test_table = Table([(1, 2, 3), (1, -1, -1)], names=('a', 'b'),
                       masked=masked)
    return test_table


def test_clean_criteria_none_removed():
    """
    If all rows satisfy the criteria, none should be removed.
    """
    inp = a_table()
    criteria = {'a': '>0'}
    out = catalog_clean(inp, **criteria)
    assert len(out) == len(inp)
    assert (out == inp).all()


@pytest.mark.parametrize("condition",
                         ['>0', '=1', '!=-1', '>=1'])
def test_clean_criteria_some_removed(condition):
    """
    Try a few filters which remove the second row and check that it is
    removed.
    """
    inp = a_table()
    criteria = {'b': condition}
    out = catalog_clean(inp, **criteria)
    assert len(out) == 1
    assert (out[0] == inp[0]).all()


@pytest.mark.parametrize("clean_masked",
                         [False, True])
def test_clean_masked_handled_correctly(clean_masked):
    inp = a_table(masked=True)
    # Mask negative values
    inp['b'].mask = inp['b'] < 0
    out = catalog_clean(inp, remove_rows_with_mask=clean_masked)
    if clean_masked:
        assert len(out) == 1
        assert (np.array(out[0]) == np.array(inp[0])).all()
    else:
        assert len(out) == len(inp)
        assert (out == inp).all()


def test_clean_masked_and_criteria():
    """
    Check whether removing masked rows and using a criteria work
    together.
    """
    inp = a_table(masked=True)
    # Mask the first row.
    inp['b'].mask = inp['b'] > 0

    inp_copy = inp.copy()
    # This should remove the third row.
    criteria = {'a': '<=2'}

    out = catalog_clean(inp, remove_rows_with_mask=True, **criteria)

    # Is only one row left?
    assert len(out) == 1

    # Is the row that is left the same as the second row of the input?
    assert (np.array(out[0]) == np.array(inp[1])).all()

    # Is the input table unchanged?
    assert (inp == inp_copy).all()


@pytest.mark.parametrize("criteria,error_msg", [
                         ({'a': '5'}, "not understood"),
                         ({'a': '<foo'}, "could not convert string")])
def test_clean_bad_criteria(criteria, error_msg):
    """
    Make sure the appropriate error is raised when bad criteria are used.
    """
    inp = a_table(masked=False)

    with pytest.raises(ValueError) as e:
        catalog_clean(inp, **criteria)
    assert error_msg in str(e.value)


def test_in_frame():
    # This wcs has the identity matrix as the coordinate transform
    wcs = make_wcs()
    # Make the image 10 x 10 pixels
    wcs.pixel_shape = (10, 10)
    coordinates_all_out = SkyCoord(ra=[50, 50], dec=[0, 0], unit='degree')
    should_all_be_out = in_frame(wcs, coordinates_all_out)
    assert not all(should_all_be_out)

    ra_cen, dec_cen = wcs.wcs.crval
    coordinates_all_in = SkyCoord(ra=[ra_cen, ra_cen + 1],
                                  dec=[dec_cen, dec_cen + 1],
                                  unit='degree')
    should_be_all_in = in_frame(wcs, coordinates_all_in)
    assert all(should_be_all_in)

    some_in_some_out = SkyCoord(ra=[50, ra_cen + 1],
                                dec=[0, dec_cen + 1],
                                unit='degree')

    in_out = in_frame(wcs, some_in_some_out)
    assert not in_out[0]
    assert in_out[1]


@pytest.mark.parametrize('clip, data_file',
                         [(True, 'data/clipped_ey_uma_vsx.fits'),
                          (False, 'data/unclipped_ey_uma_vsx.fits')])
def test_catalog_search(clip, data_file):
    # Check that a catalog search of VSX gives us what we expect.
    # I suppose this really isn't future-proof, since more variables
    # could be discovered in the future....
    data = get_pkg_data_filename(data_file)
    expected = Table.read(data)
    wcs_file = get_pkg_data_filename('data/sample_wcs_ey_uma.fits')
    wcs = WCS(fits.open(wcs_file)[0].header)
    wcs.pixel_shape = list(reversed(CCD_SHAPE))
    actual = catalog_search(wcs, CCD_SHAPE, 'B/vsx/vsx',
                            clip_by_frame=clip)
    assert all(actual['OID'] == expected['OID'])


def test_find_known_variables():
    # Under the hood this calls catalog search on the VSX
    # catalog and clips to the frame.
    data_file = 'data/clipped_ey_uma_vsx.fits'
    data = get_pkg_data_filename(data_file)
    expected = Table.read(data)
    wcs_file = get_pkg_data_filename('data/sample_wcs_ey_uma.fits')
    wcs = WCS(fits.open(wcs_file)[0].header)
    wcs.pixel_shape = list(reversed(CCD_SHAPE))
    ccd = CCDData(data=np.zeros(CCD_SHAPE), wcs=wcs, unit='adu')
    vsx = find_known_variables(ccd)
    assert expected['OID'] == vsx['OID']
    assert expected['Name'] == vsx['Name']


def test_catalog_search_from_wcs_or_coord():
    data_file = 'data/sample_wcs_ey_uma.fits'
    data = get_pkg_data_filename(data_file)
    wcs = WCS(fits.open(data)[0].header)
    wcs.pixel_shape = (4096, 4096)
    # Try the search using the WCS alone
    vsx_vars = catalog_search(wcs, [4096, 4096], 'B/vsx/vsx',
                              clip_by_frame=False)
    # And try it using a coordinate instead
    cen_coord = wcs.pixel_to_world(4096 / 2, 4096 / 2)
    # But right now clipping by frame doesn't work without the WCS
    # so don't do that
    vsx_vars2 = catalog_search(cen_coord, [4096, 4096], 'B/vsx/vsx',
                               clip_by_frame=False)
    assert len(vsx_vars) > 0
    assert len(vsx_vars) == len(vsx_vars2)


def test_catalog_search_with_coord_and_frame_clip_fails():
    # Check that calling catalog_search with a coordinate instead
    # of WCS and with clip_by_frame = True generates an appropriate
    # error.
    data_file = 'data/sample_wcs_ey_uma.fits'
    data = get_pkg_data_filename(data_file)
    wcs = WCS(fits.open(data)[0].header)
    cen_coord = wcs.pixel_to_world(4096 / 2, 4096 / 2)
    with pytest.raises(ValueError) as e:
        _ = catalog_search(cen_coord, [4096, 4096], 'B/vsx/vsx',
                           clip_by_frame=True)
    assert 'To clip entries by frame' in str(e.value)


def test_find_apass():
    # This is really checking from APASS DR9 on Vizier, or at least that
    # is where the "expected" data is drawn from.
    expected_all = Table.read(get_pkg_data_filename('data/all_apass_ey_uma_sorted_ra_first_20.fits'))
    expected_low_error = Table.read(get_pkg_data_filename('data/low_error_apass_ey_uma_sorted_ra_first_20.fits'))
    wcs_file = get_pkg_data_filename('data/sample_wcs_ey_uma.fits')
    wcs = WCS(fits.open(wcs_file)[0].header)
    wcs.pixel_shape = list(reversed(CCD_SHAPE))
    ccd = CCDData(data=np.zeros(CCD_SHAPE), wcs=wcs, unit='adu')
    all_apass, apass_low_error = find_apass_stars(ccd)
    # print(all_apass)
    # REference data was sorted by RA, first 20 entries kept
    all_apass.sort('RAJ2000')
    all_apass = all_apass[:20]
    apass_low_error.sort('RAJ2000')
    apass_low_error = apass_low_error[:20]
    # It is hard to imagine the RAs matching and other entries not matching,
    # so just check the RAs.
    assert all(all_apass['RAJ2000'] == expected_all['RAJ2000'])
    assert all(apass_low_error['RAJ2000'] == expected_low_error['RAJ2000'])


def test_filter_catalog():
    # Check basic functionality of table filtering.
    table = a_table()
    print(table)
    output = filter_catalog(table, a=1.5)
    assert output[0]
    assert not output[1]



from stellarphot.analysis.exotic import exotic_settings_widget, get_values_from_widget


def test_json_has_stuff():
    widget = exotic_settings_widget()
    stuff = get_values_from_widget(widget, key='known')
    assert 'user_info' in stuff.keys()



import re

import pytest

from stellarphot.io.tess import TessSubmission

GOOD_HEADER = {
    "date-obs": "2022-06-04T05:44:28.010",
    "filter": "ip",
    "object": "TIC-237205154"
}

GOOD_HEADER_WITH_PLANET = {
    "date-obs": "2022-06-04T05:44:28.010",
    "filter": "ip",
    "object": "TIC-237205154.01"
}

BAD_HEADER = {}


def test_good_header_sucess():
    tsub = TessSubmission.from_header(GOOD_HEADER)
    assert tsub.utc_start == "20220604"
    assert tsub.tic_id == 237205154
    assert tsub.planet_number == 0
    assert tsub.telescope_code == ""


def test_good_header_planet_sucess():
    tsub = TessSubmission.from_header(GOOD_HEADER_WITH_PLANET)
    assert tsub.utc_start == "20220604"
    assert tsub.tic_id == 237205154
    assert tsub.planet_number == 1
    assert tsub.telescope_code == ""


def test_bad_header_fails():
    with pytest.raises(ValueError) as e:
        TessSubmission.from_header(BAD_HEADER)
    match_str = ".*UTC date of first image.*filter/passband.*TIC ID number.*"
    assert re.search(match_str, str(e))


def test_base_name():
    tsub = TessSubmission.from_header(GOOD_HEADER_WITH_PLANET, telescope_code="ABS")
    assert tsub.base_name == "TIC237205154-01_20220604_ABS_ip"


def test_seeing_profile():
    tsub = TessSubmission.from_header(GOOD_HEADER_WITH_PLANET, telescope_code="ABS")
    assert tsub.seeing_profile == "TIC237205154-01_20220604_ABS_ip_seeing-profile.png"


def test_valid_method():
    tsub = TessSubmission.from_header(GOOD_HEADER)

    # Missing telescope code, amopng other things...
    assert not tsub._valid()

    # Fix the code, should still not be valid because of planet number
    tsub.telescope_code = "ABC"
    assert not tsub._valid()

    # Set planet number and should be valid
    tsub.planet_number = 1
    assert tsub._valid()

    # Set invalid TIC number
    tsub.tic_id = 10_000_000_000
    assert not tsub._valid()


# This file is used to configure the behavior of pytest when using the Astropy
# test infrastructure.
import os

from pytest_astropy_header.display import PYTEST_HEADER_MODULES, TESTED_VERSIONS


def pytest_configure(config):

    config.option.astropy_header = True
    PYTEST_HEADER_MODULES.pop('h5py', None)

    from .version import version
    packagename = os.path.basename(os.path.dirname(__file__))
    TESTED_VERSIONS[packagename] = version


# from astropy.tests.helper import enable_deprecations_as_exceptions

## Uncomment the following line to treat all DeprecationWarnings as
## exceptions. For Astropy v2.0 or later, there are 2 additional keywords,
## as follow (although default should work for most cases).
## To ignore some packages that produce deprecation warnings on import
## (in addition to 'compiler', 'scipy', 'pygments', 'ipykernel', and
## 'setuptools'), add:
##     modules_to_ignore_on_import=['module_1', 'module_2']
## To ignore some specific deprecation warning messages for Python version
## MAJOR.MINOR or later, add:
##     warnings_to_ignore_by_pyver={(MAJOR, MINOR): ['Message to ignore']}
# enable_deprecations_as_exceptions()



import numpy as np

import pytest

from photutils.datasets import make_gaussian_sources_image, make_noise_image
from astropy.table import Table
from astrowidgets import ImageWidget

from stellarphot.visualization import seeing_profile_functions as spf

# Make a few round stars
STARS = Table(dict(amplitude=[1000, 200, 300],
                   x_mean=[30, 100, 150],
                   y_mean=[40, 110, 160],
                   x_stddev=[4, 4, 4],
                   y_stddev=[4, 4, 4],
                   theta=[0, 0, 0]
                   )
)
SHAPE = (300, 300)
RANDOM_SEED = 1230971

def test_keybindings():
    def simple_bindmap(bindmap):
        bound_keys = {}
        # The keys of the event map are...messy. This converts them to strings
        for key in bindmap.keys():
            modifier = key[1]
            key_name = key[2]
            bound_keys[str(key[0]) + ''.join(modifier) + key_name] = key
        return bound_keys

    # This test assumes the ginga widget backend...
    iw = ImageWidget()
    original_bindings = iw._viewer.get_bindmap().eventmap

    bound_keys = simple_bindmap(original_bindings)
    # Spot check a couple of things before we run our function
    assert 'Nonekp_D' in bound_keys
    assert 'Nonekp_+' in bound_keys
    assert 'Nonekp_left' not in bound_keys

    # rebind
    spf.set_keybindings(iw)
    new_bindings = iw._viewer.get_bindmap().eventmap
    bound_keys = simple_bindmap(new_bindings)
    assert 'Nonekp_D' not in bound_keys
    assert 'Nonekp_+' in bound_keys
    # Yes, the line below is correct...
    assert new_bindings[bound_keys['Nonekp_left']]['name'] == 'pan_right'


def test_find_center_no_noise_good_guess():
    image = make_gaussian_sources_image(SHAPE, STARS)
    # Good initial guess, no noise, should converge in one try
    cen1 = spf.find_center(image, (31, 41), max_iters=1)
    np.testing.assert_allclose(cen1, [30, 40])


def test_find_center_noise_bad_guess():
    image = make_gaussian_sources_image(SHAPE, STARS)
    noise = make_noise_image(SHAPE, distribution='gaussian', mean=0, stddev=5,
                             seed=RANDOM_SEED)
    cen2 = spf.find_center(image + noise, [40, 50], max_iters=1)
    # Bad initial guess, noise, should take more than one try...
    with pytest.raises(AssertionError):
        np.testing.assert_allclose(cen2, [30, 40])


def test_find_center_noise_good_guess():
    image = make_gaussian_sources_image(SHAPE, STARS)
    noise = make_noise_image(SHAPE, distribution='gaussian', mean=0, stddev=5,
                             seed=RANDOM_SEED)
    # Trying again with several iterations should work
    cen3 = spf.find_center(image + noise, [31, 41], max_iters=10)
    # Tolerance chosen based on some trial and error
    np.testing.assert_allclose(cen3, [30, 40], atol=0.02)


def test_find_center_no_noise_star_at_edge():
    # Trying to put the star at the edge of the initial guess
    image = make_gaussian_sources_image(SHAPE, STARS)
    cen = spf.find_center(image, [45, 65], max_iters=10)
    np.testing.assert_allclose(cen, [30, 40])


def test_find_center_no_star():
    # No star anywhere near the original guess
    image = make_gaussian_sources_image(SHAPE, STARS)
    # Offset the mean from zero to avoid nan center
    noise = make_noise_image(SHAPE, distribution='gaussian',
                             mean=1000, stddev=5, seed=RANDOM_SEED)
    cen = spf.find_center(image + noise, [50, 200], max_iters=10)
    assert (np.abs(cen[0] - 50) > 1) and (np.abs(cen[1] - 200) > 1)


def test_radial_profile():
    image = make_gaussian_sources_image(SHAPE, STARS)
    for row in STARS:
        cen = spf.find_center(image, (row['x_mean'], row['y_mean']),
                              max_iters=10)
        print(row)
        r_ex, r_a, radprof = spf.radial_profile(image, cen)
        r_exs, r_as, radprofs = spf.radial_profile(image, cen,
                                                   return_scaled=False)

        # Numerical value below is integral of input 2D gaussian, 2pi A sigma^2
        expected_integral = 2 * np.pi * row['amplitude'] * row['x_stddev']**2
        print(expected_integral, radprofs.sum())
        np.testing.assert_allclose(radprofs.sum(), expected_integral, atol=50)
